"""
Tests for research_library/downloaders/pubmed.py

Tests cover:
- PubMedDownloader initialization
- can_handle() URL detection
- download() methods
- download_with_result() methods
- PDF download methods
- Text download methods
- Rate limiting
- PMC ID extraction
- Europe PMC API integration
- Error handling
"""

from unittest.mock import MagicMock, patch
import time


class TestPubMedDownloaderInitialization:
    """Tests for PubMedDownloader initialization."""

    def test_default_initialization(self):
        """Test default initialization parameters."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert downloader.timeout == 30
        assert downloader.rate_limit_delay == 1.0
        assert downloader.last_request_time == 0

    def test_custom_timeout(self):
        """Test initialization with custom timeout."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader(timeout=60)

        assert downloader.timeout == 60

    def test_custom_rate_limit(self):
        """Test initialization with custom rate limit."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader(rate_limit_delay=2.0)

        assert downloader.rate_limit_delay == 2.0


class TestCanHandle:
    """Tests for can_handle() URL detection."""

    def test_can_handle_pubmed_url(self):
        """Test PubMed main site URL detection."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert (
            downloader.can_handle("https://pubmed.ncbi.nlm.nih.gov/12345678")
            is True
        )
        assert (
            downloader.can_handle("https://pubmed.ncbi.nlm.nih.gov/12345678/")
            is True
        )

    def test_can_handle_pmc_url(self):
        """Test PMC URL detection."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert (
            downloader.can_handle(
                "https://ncbi.nlm.nih.gov/pmc/articles/PMC1234567"
            )
            is True
        )
        assert (
            downloader.can_handle(
                "https://ncbi.nlm.nih.gov/pmc/articles/PMC1234567/"
            )
            is True
        )

    def test_can_handle_europe_pmc_url(self):
        """Test Europe PMC URL detection."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert (
            downloader.can_handle("https://europepmc.org/article/PMC/1234567")
            is True
        )
        assert (
            downloader.can_handle(
                "https://www.europepmc.org/article/PMC/1234567"
            )
            is True
        )

    def test_cannot_handle_generic_url(self):
        """Test that generic URLs are not handled."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert downloader.can_handle("https://google.com") is False
        assert downloader.can_handle("https://arxiv.org/abs/1234") is False
        assert downloader.can_handle("https://nature.com/article/123") is False

    def test_cannot_handle_empty_url(self):
        """Test that empty URL returns False."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert downloader.can_handle("") is False

    def test_cannot_handle_invalid_url(self):
        """Test that invalid URL returns False."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert downloader.can_handle("not a valid url") is False

    def test_cannot_handle_ncbi_without_pmc(self):
        """Test that NCBI URLs without /pmc are not handled."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        # ncbi.nlm.nih.gov without /pmc should return False
        assert (
            downloader.can_handle("https://ncbi.nlm.nih.gov/gene/12345")
            is False
        )


class TestDownload:
    """Tests for download() method."""

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_apply_rate_limit",
    )
    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_pdf_content",
    )
    def test_download_pdf_success(self, mock_download_pdf, mock_rate_limit):
        """Test successful PDF download."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )
        from local_deep_research.research_library.downloaders.base import (
            ContentType,
        )

        mock_download_pdf.return_value = b"%PDF-1.4 content"

        downloader = PubMedDownloader()
        result = downloader.download(
            "https://pubmed.ncbi.nlm.nih.gov/12345678", ContentType.PDF
        )

        assert result == b"%PDF-1.4 content"
        mock_rate_limit.assert_called_once()

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_apply_rate_limit",
    )
    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_text",
    )
    def test_download_text_success(self, mock_download_text, mock_rate_limit):
        """Test successful text download."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )
        from local_deep_research.research_library.downloaders.base import (
            ContentType,
        )

        mock_download_text.return_value = b"Article text content"

        downloader = PubMedDownloader()
        result = downloader.download(
            "https://pubmed.ncbi.nlm.nih.gov/12345678", ContentType.TEXT
        )

        assert result == b"Article text content"


class TestDownloadWithResult:
    """Tests for download_with_result() method."""

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_apply_rate_limit",
    )
    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_text",
    )
    def test_download_text_with_result_success(
        self, mock_download_text, mock_rate_limit
    ):
        """Test successful text download returns success result."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )
        from local_deep_research.research_library.downloaders.base import (
            ContentType,
        )

        mock_download_text.return_value = b"Article text content"

        downloader = PubMedDownloader()
        result = downloader.download_with_result(
            "https://pubmed.ncbi.nlm.nih.gov/12345678", ContentType.TEXT
        )

        assert result.is_success is True
        assert result.content == b"Article text content"

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_apply_rate_limit",
    )
    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_text",
    )
    def test_download_text_with_result_failure(
        self, mock_download_text, mock_rate_limit
    ):
        """Test failed text download returns skip reason."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )
        from local_deep_research.research_library.downloaders.base import (
            ContentType,
        )

        mock_download_text.return_value = None

        downloader = PubMedDownloader()
        result = downloader.download_with_result(
            "https://pubmed.ncbi.nlm.nih.gov/12345678", ContentType.TEXT
        )

        assert result.is_success is False
        assert "subscription" in result.skip_reason.lower()


class TestApplyRateLimit:
    """Tests for _apply_rate_limit() method."""

    def test_no_delay_on_first_request(self):
        """Test that first request doesn't delay."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader(rate_limit_delay=1.0)
        downloader.last_request_time = 0

        start_time = time.time()
        downloader._apply_rate_limit()
        elapsed = time.time() - start_time

        # Should be nearly instant (no delay)
        assert elapsed < 0.1

    def test_delay_on_rapid_requests(self):
        """Test that rapid requests are rate limited."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader(rate_limit_delay=0.2)

        # First request
        downloader._apply_rate_limit()

        # Second request immediately after
        start_time = time.time()
        downloader._apply_rate_limit()
        elapsed = time.time() - start_time

        # Should have delayed close to rate_limit_delay
        assert elapsed >= 0.15  # Allow some tolerance

    def test_no_delay_after_waiting(self):
        """Test that there's no delay if enough time has passed."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader(rate_limit_delay=0.1)

        # Set last request time to well in the past
        downloader.last_request_time = time.time() - 10

        start_time = time.time()
        downloader._apply_rate_limit()
        elapsed = time.time() - start_time

        # Should be nearly instant
        assert elapsed < 0.05


class TestGetPmcIdFromPmid:
    """Tests for _get_pmc_id_from_pmid() method."""

    @patch("requests.Session.get")
    def test_get_pmc_id_success(self, mock_get):
        """Test successful PMC ID retrieval."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        # Mock NCBI E-utilities response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "linksets": [{"linksetdbs": [{"dbto": "pmc", "links": [7654321]}]}]
        }
        mock_get.return_value = mock_response

        downloader = PubMedDownloader()
        result = downloader._get_pmc_id_from_pmid("12345678")

        assert result == "PMC7654321"

    @patch("requests.Session.get")
    def test_get_pmc_id_no_link(self, mock_get):
        """Test when no PMC link exists."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        # Mock response with no PMC links
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"linksets": [{}]}
        mock_get.return_value = mock_response

        downloader = PubMedDownloader()
        result = downloader._get_pmc_id_from_pmid("12345678")

        assert result is None

    @patch("requests.Session.get")
    def test_get_pmc_id_api_error(self, mock_get):
        """Test PMC ID retrieval when API fails."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_get.side_effect = Exception("Network error")

        downloader = PubMedDownloader()
        result = downloader._get_pmc_id_from_pmid("12345678")

        assert result is None


class TestDownloadViaMethods:
    """Tests for _download_via_* methods."""

    @patch("requests.Session.get")
    def test_download_via_europe_pmc_success(self, mock_get):
        """Test successful download from Europe PMC."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b"%PDF-1.4 Europe PMC content"
        mock_response.headers = {"Content-Type": "application/pdf"}
        mock_get.return_value = mock_response

        downloader = PubMedDownloader()
        result = downloader._download_via_europe_pmc("PMC1234567")

        assert result == b"%PDF-1.4 Europe PMC content"

    @patch("requests.Session.get")
    def test_download_via_europe_pmc_failure(self, mock_get):
        """Test failed download from Europe PMC."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_get.return_value = mock_response

        downloader = PubMedDownloader()
        result = downloader._download_via_europe_pmc("PMC1234567")

        assert result is None

    @patch("requests.Session.get")
    def test_download_via_ncbi_pmc_success(self, mock_get):
        """Test successful download from NCBI PMC."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b"%PDF-1.4 NCBI PMC content"
        mock_response.headers = {"Content-Type": "application/pdf"}
        mock_get.return_value = mock_response

        downloader = PubMedDownloader()
        result = downloader._download_via_ncbi_pmc("PMC1234567")

        assert result == b"%PDF-1.4 NCBI PMC content"


class TestDownloadPmcDirect:
    """Tests for _download_pmc_direct() method."""

    def test_extract_pmc_id_from_url(self):
        """Test PMC ID extraction from URL."""
        import re

        url = "https://ncbi.nlm.nih.gov/pmc/articles/PMC7654321"
        pmc_match = re.search(r"(PMC\d+)", url)

        assert pmc_match is not None
        assert pmc_match.group(1) == "PMC7654321"

    def test_pmc_id_not_found(self):
        """Test when PMC ID is not in URL."""
        import re

        url = "https://ncbi.nlm.nih.gov/pmc/articles/"
        pmc_match = re.search(r"(PMC\d+)", url)

        assert pmc_match is None


class TestDownloadPubmed:
    """Tests for _download_pubmed() method."""

    def test_extract_pmid_from_url(self):
        """Test PMID extraction from URL."""
        import re

        url = "https://pubmed.ncbi.nlm.nih.gov/12345678/"
        pmid_match = re.search(r"/(\d+)/?", url)

        assert pmid_match is not None
        assert pmid_match.group(1) == "12345678"

    def test_pmid_not_found(self):
        """Test when PMID is not in URL."""
        import re

        url = "https://pubmed.ncbi.nlm.nih.gov/"
        pmid_match = re.search(r"/(\d+)/?", url)

        assert pmid_match is None


class TestTryEuropePmcApi:
    """Tests for _try_europe_pmc_api() method."""

    @patch("requests.Session.get")
    def test_api_returns_open_access(self, mock_get):
        """Test API returns open access article."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        # First call - API search
        api_response = MagicMock()
        api_response.status_code = 200
        api_response.json.return_value = {
            "resultList": {
                "result": [
                    {
                        "isOpenAccess": "Y",
                        "hasPDF": "Y",
                        "pmcid": "PMC7654321",
                    }
                ]
            }
        }

        # Second call - PDF download
        pdf_response = MagicMock()
        pdf_response.status_code = 200
        pdf_response.content = b"%PDF-1.4 content"
        pdf_response.headers = {"Content-Type": "application/pdf"}

        mock_get.side_effect = [api_response, pdf_response]

        downloader = PubMedDownloader()
        result = downloader._try_europe_pmc_api("12345678")

        assert result == b"%PDF-1.4 content"

    @patch("requests.Session.get")
    def test_api_returns_no_results(self, mock_get):
        """Test API returns no results."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"resultList": {"result": []}}
        mock_get.return_value = mock_response

        downloader = PubMedDownloader()
        result = downloader._try_europe_pmc_api("12345678")

        assert result is None

    @patch("requests.Session.get")
    def test_api_returns_non_open_access(self, mock_get):
        """Test API returns non-open access article."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "resultList": {
                "result": [
                    {
                        "isOpenAccess": "N",
                        "hasPDF": "N",
                    }
                ]
            }
        }
        mock_get.return_value = mock_response

        downloader = PubMedDownloader()
        result = downloader._try_europe_pmc_api("12345678")

        assert result is None


class TestDownloadPdfWithResult:
    """Tests for _download_pdf_with_result() method."""

    def test_invalid_pmc_url_format(self):
        """Test invalid PMC URL format returns error result."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()
        downloader._apply_rate_limit = MagicMock()  # Skip rate limiting

        result = downloader._download_pdf_with_result(
            "https://ncbi.nlm.nih.gov/pmc/articles/"
        )

        # Should return skip reason about invalid format
        assert result.is_success is False
        assert result.skip_reason is not None

    def test_invalid_pubmed_url_format(self):
        """Test invalid PubMed URL format returns error result."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()
        downloader._apply_rate_limit = MagicMock()

        result = downloader._download_pdf_with_result(
            "https://pubmed.ncbi.nlm.nih.gov/"
        )

        assert result.is_success is False
        assert result.skip_reason is not None


class TestFetchTextFromEuropePmc:
    """Tests for _fetch_text_from_europe_pmc() method."""

    @patch("requests.Session.get")
    def test_fetch_text_success(self, mock_get):
        """Test successful text fetch from Europe PMC."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        # First call - metadata
        meta_response = MagicMock()
        meta_response.status_code = 200
        meta_response.json.return_value = {
            "resultList": {
                "result": [
                    {
                        "isOpenAccess": "Y",
                        "pmcid": "PMC7654321",
                    }
                ]
            }
        }

        # Second call - full text XML
        xml_response = MagicMock()
        xml_response.status_code = 200
        xml_response.text = (
            "<article><body><p>Article text content</p></body></article>"
        )

        mock_get.side_effect = [meta_response, xml_response]

        downloader = PubMedDownloader()
        result = downloader._fetch_text_from_europe_pmc("12345678", None)

        assert result is not None
        assert "Article text content" in result

    @patch("requests.Session.get")
    def test_fetch_text_no_open_access(self, mock_get):
        """Test text fetch when article is not open access."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "resultList": {
                "result": [
                    {
                        "isOpenAccess": "N",
                    }
                ]
            }
        }
        mock_get.return_value = mock_response

        downloader = PubMedDownloader()
        result = downloader._fetch_text_from_europe_pmc("12345678", None)

        assert result is None

    def test_fetch_text_no_identifiers(self):
        """Test text fetch with no identifiers."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()
        result = downloader._fetch_text_from_europe_pmc(None, None)

        assert result is None


class TestDownloadText:
    """Tests for _download_text() method."""

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_fetch_text_from_europe_pmc",
    )
    def test_download_text_from_pubmed_url(self, mock_fetch_text):
        """Test text download from PubMed URL."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_fetch_text.return_value = "Full article text"

        downloader = PubMedDownloader()
        result = downloader._download_text(
            "https://pubmed.ncbi.nlm.nih.gov/12345678/"
        )

        assert result == b"Full article text"
        mock_fetch_text.assert_called_once()

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_fetch_text_from_europe_pmc",
    )
    def test_download_text_from_pmc_url(self, mock_fetch_text):
        """Test text download from PMC URL."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_fetch_text.return_value = "PMC article text"

        downloader = PubMedDownloader()
        result = downloader._download_text(
            "https://ncbi.nlm.nih.gov/pmc/articles/PMC7654321/"
        )

        assert result == b"PMC article text"


class TestDownloadPdfContent:
    """Tests for _download_pdf_content() method."""

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_pmc_direct",
    )
    def test_routes_pmc_url_correctly(self, mock_download_pmc):
        """Test that PMC URLs are routed to _download_pmc_direct."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_download_pmc.return_value = b"%PDF-1.4 content"

        downloader = PubMedDownloader()
        result = downloader._download_pdf_content(
            "https://ncbi.nlm.nih.gov/pmc/articles/PMC7654321"
        )

        mock_download_pmc.assert_called_once()
        assert result == b"%PDF-1.4 content"

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_pubmed",
    )
    def test_routes_pubmed_url_correctly(self, mock_download_pubmed):
        """Test that PubMed URLs are routed to _download_pubmed."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_download_pubmed.return_value = b"%PDF-1.4 content"

        downloader = PubMedDownloader()
        result = downloader._download_pdf_content(
            "https://pubmed.ncbi.nlm.nih.gov/12345678"
        )

        mock_download_pubmed.assert_called_once()
        assert result == b"%PDF-1.4 content"

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_europe_pmc",
    )
    def test_routes_europe_pmc_url_correctly(self, mock_download_europe):
        """Test that Europe PMC URLs are routed correctly."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_download_europe.return_value = b"%PDF-1.4 content"

        downloader = PubMedDownloader()
        result = downloader._download_pdf_content(
            "https://europepmc.org/article/PMC/7654321"
        )

        mock_download_europe.assert_called_once()
        assert result == b"%PDF-1.4 content"


class TestDownloadEuropePmc:
    """Tests for _download_europe_pmc() method."""

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_via_europe_pmc",
    )
    def test_extracts_pmc_id_and_downloads(self, mock_download):
        """Test PMC ID extraction and download."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        mock_download.return_value = b"%PDF-1.4 content"

        downloader = PubMedDownloader()
        result = downloader._download_europe_pmc(
            "https://europepmc.org/article/PMC7654321"
        )

        mock_download.assert_called_once_with("PMC7654321")
        assert result == b"%PDF-1.4 content"

    @patch.object(
        __import__(
            "local_deep_research.research_library.downloaders.pubmed",
            fromlist=["PubMedDownloader"],
        ).PubMedDownloader,
        "_download_via_europe_pmc",
    )
    def test_returns_none_when_no_pmc_id(self, mock_download):
        """Test returns None when PMC ID not found."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()
        result = downloader._download_europe_pmc(
            "https://europepmc.org/article/invalid"
        )

        mock_download.assert_not_called()
        assert result is None


class TestEdgeCases:
    """Edge case tests."""

    def test_url_with_query_parameters(self):
        """Test URL with query parameters."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        # Should still handle URLs with query params
        assert (
            downloader.can_handle(
                "https://pubmed.ncbi.nlm.nih.gov/12345678?from=home"
            )
            is True
        )

    def test_url_with_fragment(self):
        """Test URL with fragment."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert (
            downloader.can_handle(
                "https://pubmed.ncbi.nlm.nih.gov/12345678#abstract"
            )
            is True
        )

    def test_http_url(self):
        """Test HTTP (non-HTTPS) URL."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        # Should handle HTTP URLs too
        assert (
            downloader.can_handle("http://pubmed.ncbi.nlm.nih.gov/12345678")
            is True
        )

    def test_url_parsing_exception(self):
        """Test URL that causes parsing exception."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        # Invalid URL should return False without raising
        result = downloader.can_handle("://invalid")
        assert result is False


class TestBaseDownloaderInheritance:
    """Tests for BaseDownloader inheritance."""

    def test_inherits_from_base_downloader(self):
        """Test that PubMedDownloader inherits from BaseDownloader."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )
        from local_deep_research.research_library.downloaders.base import (
            BaseDownloader,
        )

        downloader = PubMedDownloader()

        assert isinstance(downloader, BaseDownloader)

    def test_has_session_attribute(self):
        """Test that downloader has session attribute."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert hasattr(downloader, "session")

    def test_has_download_pdf_method(self):
        """Test that downloader has _download_pdf method from base."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert hasattr(downloader, "_download_pdf")
        assert callable(downloader._download_pdf)

    def test_has_extract_text_from_pdf_method(self):
        """Test that downloader has extract_text_from_pdf method from base."""
        from local_deep_research.research_library.downloaders.pubmed import (
            PubMedDownloader,
        )

        downloader = PubMedDownloader()

        assert hasattr(downloader, "extract_text_from_pdf")
        assert callable(downloader.extract_text_from_pdf)



"""
Tests for benchmarks/metrics/calculation.py

Tests cover:
- calculate_metrics function
- calculate_combined_score function
- calculate_resource_metrics function
"""

import json


class TestCalculateMetrics:
    """Tests for the calculate_metrics function."""

    def test_calculates_basic_metrics(self, tmp_path):
        """Test calculation of basic metrics from results file."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_metrics,
        )

        # Create a test results file
        results_file = tmp_path / "results.jsonl"
        results = [
            {"is_correct": True, "processing_time": 1.5, "confidence": "90"},
            {"is_correct": True, "processing_time": 2.0, "confidence": "85"},
            {"is_correct": False, "processing_time": 1.0, "confidence": "70"},
        ]
        with open(results_file, "w") as f:
            for r in results:
                f.write(json.dumps(r) + "\n")

        metrics = calculate_metrics(str(results_file))

        assert metrics["total_examples"] == 3
        assert metrics["graded_examples"] == 3
        assert metrics["correct"] == 2
        assert metrics["accuracy"] == 2 / 3
        assert metrics["average_processing_time"] == 1.5
        assert metrics["average_confidence"] == (90 + 85 + 70) / 3

    def test_handles_empty_file(self, tmp_path):
        """Test handling of empty results file."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_metrics,
        )

        results_file = tmp_path / "empty.jsonl"
        results_file.write_text("")

        metrics = calculate_metrics(str(results_file))

        assert "error" in metrics
        assert metrics["error"] == "No results found"

    def test_handles_missing_file(self, tmp_path):
        """Test handling of missing results file."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_metrics,
        )

        metrics = calculate_metrics(str(tmp_path / "nonexistent.jsonl"))

        assert "error" in metrics

    def test_calculates_error_rate(self, tmp_path):
        """Test calculation of error rate."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_metrics,
        )

        results_file = tmp_path / "results.jsonl"
        results = [
            {"is_correct": True},
            {"error": "Some error occurred"},
            {"is_correct": False},
        ]
        with open(results_file, "w") as f:
            for r in results:
                f.write(json.dumps(r) + "\n")

        metrics = calculate_metrics(str(results_file))

        assert metrics["error_count"] == 1
        assert metrics["error_rate"] == 1 / 3

    def test_calculates_per_category_metrics(self, tmp_path):
        """Test calculation of per-category metrics."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_metrics,
        )

        results_file = tmp_path / "results.jsonl"
        results = [
            {"is_correct": True, "category": "science"},
            {"is_correct": False, "category": "science"},
            {"is_correct": True, "category": "history"},
            {"is_correct": True, "category": "history"},
        ]
        with open(results_file, "w") as f:
            for r in results:
                f.write(json.dumps(r) + "\n")

        metrics = calculate_metrics(str(results_file))

        assert "categories" in metrics
        assert metrics["categories"]["science"]["total"] == 2
        assert metrics["categories"]["science"]["correct"] == 1
        assert metrics["categories"]["science"]["accuracy"] == 0.5
        assert metrics["categories"]["history"]["total"] == 2
        assert metrics["categories"]["history"]["correct"] == 2
        assert metrics["categories"]["history"]["accuracy"] == 1.0

    def test_handles_missing_confidence(self, tmp_path):
        """Test handling of results without confidence values."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_metrics,
        )

        results_file = tmp_path / "results.jsonl"
        results = [
            {"is_correct": True},
            {"is_correct": False},
        ]
        with open(results_file, "w") as f:
            for r in results:
                f.write(json.dumps(r) + "\n")

        metrics = calculate_metrics(str(results_file))

        assert metrics["average_confidence"] == 0

    def test_handles_invalid_confidence(self, tmp_path):
        """Test handling of invalid confidence values."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_metrics,
        )

        results_file = tmp_path / "results.jsonl"
        results = [
            {"is_correct": True, "confidence": "invalid"},
            {"is_correct": True, "confidence": "80"},
        ]
        with open(results_file, "w") as f:
            for r in results:
                f.write(json.dumps(r) + "\n")

        metrics = calculate_metrics(str(results_file))

        # Should skip invalid and only use valid confidence
        assert metrics["average_confidence"] == 80


class TestCalculateCombinedScore:
    """Tests for the calculate_combined_score function."""

    def test_calculates_with_default_weights(self):
        """Test combined score with default weights."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_combined_score,
        )

        metrics = {
            "quality": {"quality_score": 0.8},
            "speed": {"speed_score": 0.6},
            "resource": {"resource_score": 0.9},
        }

        score = calculate_combined_score(metrics)

        # Default weights: quality=0.6, speed=0.3, resource=0.1
        expected = 0.8 * 0.6 + 0.6 * 0.3 + 0.9 * 0.1
        assert abs(score - expected) < 0.001

    def test_calculates_with_custom_weights(self):
        """Test combined score with custom weights."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_combined_score,
        )

        metrics = {
            "quality": {"quality_score": 1.0},
            "speed": {"speed_score": 0.5},
            "resource": {"resource_score": 0.0},
        }
        weights = {"quality": 0.5, "speed": 0.5, "resource": 0.0}

        score = calculate_combined_score(metrics, weights)

        expected = 1.0 * 0.5 + 0.5 * 0.5 + 0.0 * 0.0
        assert abs(score - expected) < 0.001

    def test_handles_missing_metrics(self):
        """Test handling of missing metric categories."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_combined_score,
        )

        metrics = {
            "quality": {"quality_score": 0.8},
            # Missing speed and resource
        }

        score = calculate_combined_score(metrics)

        # Should only count quality component
        assert score > 0
        assert score <= 1.0

    def test_handles_zero_weights(self):
        """Test handling of zero total weight."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_combined_score,
        )

        metrics = {
            "quality": {"quality_score": 0.8},
        }
        weights = {"quality": 0, "speed": 0, "resource": 0}

        score = calculate_combined_score(metrics, weights)

        assert score == 0.0

    def test_normalizes_weights(self):
        """Test that weights are normalized."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_combined_score,
        )

        metrics = {
            "quality": {"quality_score": 1.0},
            "speed": {"speed_score": 1.0},
            "resource": {"resource_score": 1.0},
        }
        # Weights that don't sum to 1
        weights = {"quality": 2, "speed": 1, "resource": 1}

        score = calculate_combined_score(metrics, weights)

        # All scores are 1.0, so normalized result should be 1.0
        assert abs(score - 1.0) < 0.001


class TestCalculateResourceMetrics:
    """Tests for the calculate_resource_metrics function."""

    def test_calculates_resource_score(self):
        """Test calculation of resource score."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_resource_metrics,
        )

        config = {
            "iterations": 2,
            "questions_per_iteration": 2,
            "max_results": 50,
        }

        metrics = calculate_resource_metrics(config)

        assert "resource_score" in metrics
        assert "estimated_complexity" in metrics
        assert 0 <= metrics["resource_score"] <= 1

    def test_higher_complexity_lower_score(self):
        """Test that higher complexity results in lower resource score."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_resource_metrics,
        )

        low_complexity_config = {
            "iterations": 1,
            "questions_per_iteration": 1,
            "max_results": 25,
        }
        high_complexity_config = {
            "iterations": 5,
            "questions_per_iteration": 5,
            "max_results": 100,
        }

        low_metrics = calculate_resource_metrics(low_complexity_config)
        high_metrics = calculate_resource_metrics(high_complexity_config)

        assert low_metrics["resource_score"] > high_metrics["resource_score"]
        assert (
            low_metrics["estimated_complexity"]
            < high_metrics["estimated_complexity"]
        )

    def test_uses_default_values(self):
        """Test that default values are used for missing config."""
        from local_deep_research.benchmarks.metrics.calculation import (
            calculate_resource_metrics,
        )

        empty_config = {}

        metrics = calculate_resource_metrics(empty_config)

        assert "resource_score" in metrics
        assert metrics["resource_score"] > 0



"""
Tests for advanced_search_system/filters/followup_relevance_filter.py

Tests cover:
- FollowUpRelevanceFilter initialization
- filter_results method
- _select_relevant_sources method
"""

from unittest.mock import Mock


class TestFollowUpRelevanceFilterInit:
    """Tests for FollowUpRelevanceFilter initialization."""

    def test_inherits_from_base_filter(self):
        """Test that FollowUpRelevanceFilter inherits from BaseFilter."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )
        from local_deep_research.advanced_search_system.filters.base_filter import (
            BaseFilter,
        )

        assert issubclass(FollowUpRelevanceFilter, BaseFilter)

    def test_init_with_model(self):
        """Test initialization with a model."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        filter_obj = FollowUpRelevanceFilter(model=mock_model)

        assert filter_obj.model is mock_model

    def test_init_without_model(self):
        """Test initialization without a model."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        filter_obj = FollowUpRelevanceFilter()

        assert filter_obj.model is None


class TestFilterResults:
    """Tests for FollowUpRelevanceFilter.filter_results method."""

    def test_empty_results_returns_empty_list(self):
        """Test that empty results returns empty list."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        filter_obj = FollowUpRelevanceFilter(model=Mock())

        result = filter_obj.filter_results([], "test query")

        assert result == []

    def test_filter_results_with_model(self):
        """Test filtering results with model."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0, 2]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        results = [
            {"title": "Result 1", "url": "http://example.com/1"},
            {"title": "Result 2", "url": "http://example.com/2"},
            {"title": "Result 3", "url": "http://example.com/3"},
        ]

        filtered = filter_obj.filter_results(results, "test query")

        assert len(filtered) == 2
        assert filtered[0]["title"] == "Result 1"
        assert filtered[1]["title"] == "Result 3"

    def test_filter_results_respects_max_results(self):
        """Test that max_results parameter is passed to selection."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0, 1, 2, 3, 4]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        results = [{"title": f"Result {i}"} for i in range(10)]

        filtered = filter_obj.filter_results(results, "query", max_results=5)

        assert len(filtered) == 5

    def test_filter_results_with_past_findings(self):
        """Test filtering with past findings context."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[1]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        results = [
            {"title": "Result 1"},
            {"title": "Result 2"},
        ]

        filter_obj.filter_results(
            results, "query", past_findings="Previous research summary"
        )

        # Check that the prompt included past findings
        call_args = mock_model.invoke.call_args[0][0]
        assert "Previous research" in call_args

    def test_filter_results_with_original_query(self):
        """Test filtering with original query context."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        results = [{"title": "Result 1"}]

        filter_obj.filter_results(
            results, "follow-up query", original_query="original research query"
        )

        call_args = mock_model.invoke.call_args[0][0]
        assert "original research query" in call_args


class TestSelectRelevantSources:
    """Tests for FollowUpRelevanceFilter._select_relevant_sources method."""

    def test_no_model_returns_first_max_results(self):
        """Test that no model returns first max_results indices."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        filter_obj = FollowUpRelevanceFilter(model=None)
        sources = [{"title": f"Source {i}"} for i in range(10)]

        indices = filter_obj._select_relevant_sources(
            sources, "query", "", max_results=5
        )

        assert indices == [0, 1, 2, 3, 4]

    def test_model_returns_parsed_indices(self):
        """Test that model response is parsed correctly."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[1, 3, 5]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": f"Source {i}"} for i in range(10)]

        indices = filter_obj._select_relevant_sources(
            sources, "query", "", max_results=5
        )

        assert indices == [1, 3, 5]

    def test_handles_json_with_think_tags(self):
        """Test that think tags are removed from response."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(
            content="<think>reasoning</think>[0, 2]"
        )

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": f"Source {i}"} for i in range(5)]

        indices = filter_obj._select_relevant_sources(
            sources, "query", "", max_results=3
        )

        assert indices == [0, 2]

    def test_handles_invalid_json_with_regex_fallback(self):
        """Test that invalid JSON falls back to regex extraction."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(
            content="I recommend sources 0, 2, and 4 as most relevant."
        )

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": f"Source {i}"} for i in range(10)]

        indices = filter_obj._select_relevant_sources(
            sources, "query", "", max_results=5
        )

        assert 0 in indices
        assert 2 in indices
        assert 4 in indices

    def test_filters_out_indices_beyond_source_length(self):
        """Test that indices beyond source list are filtered out."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0, 1, 100, 200]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": f"Source {i}"} for i in range(5)]

        indices = filter_obj._select_relevant_sources(
            sources, "query", "", max_results=5
        )

        assert indices == [0, 1]

    def test_handles_model_exception(self):
        """Test that model exception falls back to first max_results."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.side_effect = Exception("Model error")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": f"Source {i}"} for i in range(10)]

        indices = filter_obj._select_relevant_sources(
            sources, "query", "", max_results=3
        )

        assert indices == [0, 1, 2]

    def test_builds_source_list_with_title_url_snippet(self):
        """Test that source list includes title, url, and snippet."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [
            {
                "title": "Test Title",
                "url": "http://example.com",
                "snippet": "Test snippet content",
            }
        ]

        filter_obj._select_relevant_sources(sources, "query", "", max_results=1)

        call_args = mock_model.invoke.call_args[0][0]
        assert "Test Title" in call_args
        assert "http://example.com" in call_args
        assert "Test snippet" in call_args

    def test_uses_content_preview_when_no_snippet(self):
        """Test that content_preview is used when snippet is missing."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [
            {
                "title": "Test",
                "content_preview": "Preview content here",
            }
        ]

        filter_obj._select_relevant_sources(sources, "query", "", max_results=1)

        call_args = mock_model.invoke.call_args[0][0]
        assert "Preview content" in call_args

    def test_truncates_long_snippets(self):
        """Test that long snippets are truncated to approximately 150 chars."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [
            {
                "title": "Test",
                "snippet": "A" * 200,
            }
        ]

        filter_obj._select_relevant_sources(sources, "query", "", max_results=1)

        call_args = mock_model.invoke.call_args[0][0]
        # Check that the snippet was truncated (not all 200 A's present)
        # Code does [:150] which results in roughly 150 chars
        assert call_args.count("A") < 200  # Definitely truncated from original

    def test_handles_float_indices(self):
        """Test that float indices are converted to int."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0.0, 2.0, 4.0]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": f"Source {i}"} for i in range(10)]

        indices = filter_obj._select_relevant_sources(
            sources, "query", "", max_results=5
        )

        assert indices == [0, 2, 4]

    def test_handles_response_not_a_list(self):
        """Test handling when response is not a list."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content='{"result": [0, 1]}')

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": f"Source {i}"} for i in range(5)]

        # Should fall back to regex extraction
        indices = filter_obj._select_relevant_sources(
            sources, "query", "", max_results=3
        )

        assert 0 in indices
        assert 1 in indices


class TestContextInPrompt:
    """Tests for context inclusion in prompts."""

    def test_includes_original_query_in_context(self):
        """Test that original query is included in context section."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": "Source 1"}]

        filter_obj._select_relevant_sources(
            sources,
            "follow-up query",
            "",
            max_results=1,
            original_query="original",
        )

        call_args = mock_model.invoke.call_args[0][0]
        assert "Original research question: original" in call_args

    def test_includes_past_findings_in_context(self):
        """Test that past findings are included in context section."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": "Source 1"}]

        filter_obj._select_relevant_sources(
            sources, "query", "Previous findings summary", max_results=1
        )

        call_args = mock_model.invoke.call_args[0][0]
        assert "Previous findings summary" in call_args

    def test_no_context_section_when_empty(self):
        """Test that context section is empty when no context provided."""
        from local_deep_research.advanced_search_system.filters.followup_relevance_filter import (
            FollowUpRelevanceFilter,
        )

        mock_model = Mock()
        mock_model.invoke.return_value = Mock(content="[0]")

        filter_obj = FollowUpRelevanceFilter(model=mock_model)
        sources = [{"title": "Source 1"}]

        filter_obj._select_relevant_sources(
            sources, "query", "", max_results=1, original_query=""
        )

        call_args = mock_model.invoke.call_args[0][0]
        assert "Previous Research Context" not in call_args



"""Tests for benchmark API functions."""

from unittest.mock import patch


class TestGetAvailableBenchmarks:
    def test_returns_list(self):
        from local_deep_research.api.benchmark_functions import (
            get_available_benchmarks,
        )

        result = get_available_benchmarks()
        assert isinstance(result, list)

    def test_contains_simpleqa(self):
        from local_deep_research.api.benchmark_functions import (
            get_available_benchmarks,
        )

        result = get_available_benchmarks()
        ids = [b["id"] for b in result]
        assert "simpleqa" in ids

    def test_contains_browsecomp(self):
        from local_deep_research.api.benchmark_functions import (
            get_available_benchmarks,
        )

        result = get_available_benchmarks()
        ids = [b["id"] for b in result]
        assert "browsecomp" in ids

    def test_contains_xbench(self):
        from local_deep_research.api.benchmark_functions import (
            get_available_benchmarks,
        )

        result = get_available_benchmarks()
        ids = [b["id"] for b in result]
        assert "xbench_deepsearch" in ids

    def test_benchmarks_have_required_fields(self):
        from local_deep_research.api.benchmark_functions import (
            get_available_benchmarks,
        )

        result = get_available_benchmarks()
        for benchmark in result:
            assert "id" in benchmark
            assert "name" in benchmark
            assert "description" in benchmark
            assert "recommended_examples" in benchmark


class TestEvaluateSimpleqa:
    def test_calls_run_benchmark(self):
        from local_deep_research.api.benchmark_functions import (
            evaluate_simpleqa,
        )

        with patch(
            "local_deep_research.api.benchmark_functions.run_simpleqa_benchmark"
        ) as mock:
            mock.return_value = {"status": "complete"}
            result = evaluate_simpleqa(num_examples=10)
            mock.assert_called_once()
            assert result["status"] == "complete"

    def test_passes_search_config(self):
        from local_deep_research.api.benchmark_functions import (
            evaluate_simpleqa,
        )

        with patch(
            "local_deep_research.api.benchmark_functions.run_simpleqa_benchmark"
        ) as mock:
            mock.return_value = {}
            evaluate_simpleqa(search_iterations=5, questions_per_iteration=7)
            call_args = mock.call_args
            assert call_args.kwargs["search_config"]["iterations"] == 5
            assert (
                call_args.kwargs["search_config"]["questions_per_iteration"]
                == 7
            )

    def test_passes_evaluation_config(self):
        from local_deep_research.api.benchmark_functions import (
            evaluate_simpleqa,
        )

        with patch(
            "local_deep_research.api.benchmark_functions.run_simpleqa_benchmark"
        ) as mock:
            mock.return_value = {}
            evaluate_simpleqa(
                evaluation_model="gpt-4", evaluation_provider="openai"
            )
            call_args = mock.call_args
            assert (
                call_args.kwargs["evaluation_config"]["model_name"] == "gpt-4"
            )
            assert call_args.kwargs["evaluation_config"]["provider"] == "openai"


class TestEvaluateBrowsecomp:
    def test_calls_run_benchmark(self):
        from local_deep_research.api.benchmark_functions import (
            evaluate_browsecomp,
        )

        with patch(
            "local_deep_research.api.benchmark_functions.run_browsecomp_benchmark"
        ) as mock:
            mock.return_value = {"status": "complete"}
            result = evaluate_browsecomp(num_examples=10)
            mock.assert_called_once()
            assert result["status"] == "complete"


class TestEvaluateXbenchDeepsearch:
    def test_calls_run_benchmark(self):
        from local_deep_research.api.benchmark_functions import (
            evaluate_xbench_deepsearch,
        )

        with patch(
            "local_deep_research.api.benchmark_functions.run_xbench_deepsearch_benchmark"
        ) as mock:
            mock.return_value = {"status": "complete"}
            result = evaluate_xbench_deepsearch(num_examples=10)
            mock.assert_called_once()
            assert result["status"] == "complete"


class TestCompareConfigurations:
    def test_returns_dict(self):
        from local_deep_research.api.benchmark_functions import (
            compare_configurations,
        )

        with patch(
            "local_deep_research.api.benchmark_functions.run_benchmark"
        ) as mock_run:
            with patch(
                "local_deep_research.security.file_write_verifier.write_file_verified"
            ):
                mock_run.return_value = {"metrics": {"accuracy": 0.8}}
                result = compare_configurations(num_examples=5)
                assert isinstance(result, dict)
                assert "status" in result

    def test_uses_default_configurations(self):
        from local_deep_research.api.benchmark_functions import (
            compare_configurations,
        )

        with patch(
            "local_deep_research.api.benchmark_functions.run_benchmark"
        ) as mock_run:
            with patch(
                "local_deep_research.security.file_write_verifier.write_file_verified"
            ):
                mock_run.return_value = {"metrics": {}}
                result = compare_configurations(num_examples=5)
                assert result["configurations_tested"] == 3

    def test_custom_configurations(self):
        from local_deep_research.api.benchmark_functions import (
            compare_configurations,
        )

        custom_configs = [
            {"name": "Custom", "search_tool": "wikipedia", "iterations": 2}
        ]
        with patch(
            "local_deep_research.api.benchmark_functions.run_benchmark"
        ) as mock_run:
            with patch(
                "local_deep_research.security.file_write_verifier.write_file_verified"
            ):
                mock_run.return_value = {"metrics": {}}
                result = compare_configurations(
                    num_examples=5, configurations=custom_configs
                )
                assert result["configurations_tested"] == 1



"""Tests for llm_config module."""

from unittest.mock import MagicMock, patch


from local_deep_research.config.llm_config import (
    VALID_PROVIDERS,
    get_setting_from_snapshot,
    is_openai_available,
    is_anthropic_available,
    is_openai_endpoint_available,
    is_ollama_available,
    is_lmstudio_available,
    is_llamacpp_available,
    is_google_available,
    is_openrouter_available,
    get_available_providers,
    get_selected_llm_provider,
    get_fallback_model,
    wrap_llm_without_think_tags,
    get_llm,
)


class TestValidProviders:
    """Tests for VALID_PROVIDERS constant."""

    def test_contains_expected_providers(self):
        """Should contain all expected providers."""
        expected = [
            "ollama",
            "openai",
            "anthropic",
            "google",
            "openrouter",
            "vllm",
            "openai_endpoint",
            "lmstudio",
            "llamacpp",
            "none",
        ]
        for provider in expected:
            assert provider in VALID_PROVIDERS

    def test_is_list(self):
        """Should be a list."""
        assert isinstance(VALID_PROVIDERS, list)


class TestGetSettingFromSnapshot:
    """Tests for get_setting_from_snapshot wrapper."""

    def test_enables_fallback_llm_check(self):
        """Should enable check_fallback_llm via get_llm_setting_from_snapshot."""
        with patch(
            "local_deep_research.config.thread_settings.get_setting_from_snapshot"
        ) as mock:
            mock.return_value = "value"
            get_setting_from_snapshot("key", "default")
            mock.assert_called_with(
                "key", "default", None, None, check_fallback_llm=True
            )


class TestIsOpenaiAvailable:
    """Tests for is_openai_available function (delegates to OpenAIProvider)."""

    def test_returns_true_when_api_key_set(self):
        """Should return True when API key is configured."""
        with patch(
            "local_deep_research.llm.providers.implementations.openai.OpenAIProvider.is_available",
            return_value=True,
        ):
            assert is_openai_available() is True

    def test_returns_false_when_no_api_key(self):
        """Should return False when no API key."""
        with patch(
            "local_deep_research.llm.providers.implementations.openai.OpenAIProvider.is_available",
            return_value=False,
        ):
            assert is_openai_available() is False

    def test_returns_false_on_exception(self):
        """Should return False on exception."""
        with patch(
            "local_deep_research.llm.providers.implementations.openai.OpenAIProvider.is_available",
            side_effect=Exception("error"),
        ):
            assert is_openai_available() is False


class TestIsAnthropicAvailable:
    """Tests for is_anthropic_available function (delegates to AnthropicProvider)."""

    def test_returns_true_when_api_key_set(self):
        """Should return True when API key is configured."""
        with patch(
            "local_deep_research.llm.providers.implementations.anthropic.AnthropicProvider.is_available",
            return_value=True,
        ):
            assert is_anthropic_available() is True

    def test_returns_false_when_no_api_key(self):
        """Should return False when no API key."""
        with patch(
            "local_deep_research.llm.providers.implementations.anthropic.AnthropicProvider.is_available",
            return_value=False,
        ):
            assert is_anthropic_available() is False


class TestIsOpenaiEndpointAvailable:
    """Tests for is_openai_endpoint_available function (delegates to CustomOpenAIEndpointProvider)."""

    def test_returns_true_when_api_key_set(self):
        """Should return True when API key is configured."""
        with patch(
            "local_deep_research.llm.providers.openai_base.OpenAICompatibleProvider.is_available",
            return_value=True,
        ):
            assert is_openai_endpoint_available() is True

    def test_returns_false_when_no_api_key(self):
        """Should return False when no API key."""
        with patch(
            "local_deep_research.llm.providers.openai_base.OpenAICompatibleProvider.is_available",
            return_value=False,
        ):
            assert is_openai_endpoint_available() is False


class TestIsOllamaAvailable:
    """Tests for is_ollama_available function (delegates to OllamaProvider)."""

    def test_returns_true_when_ollama_responds(self):
        """Should return True when OllamaProvider reports available."""
        with patch(
            "local_deep_research.llm.providers.implementations.ollama.OllamaProvider.is_available",
            return_value=True,
        ):
            assert is_ollama_available() is True

    def test_returns_false_when_ollama_not_running(self):
        """Should return False when OllamaProvider reports unavailable."""
        with patch(
            "local_deep_research.llm.providers.implementations.ollama.OllamaProvider.is_available",
            return_value=False,
        ):
            assert is_ollama_available() is False

    def test_returns_false_on_exception(self):
        """Should return False on exception."""
        with patch(
            "local_deep_research.llm.providers.implementations.ollama.OllamaProvider.is_available",
            side_effect=Exception("error"),
        ):
            assert is_ollama_available() is False


class TestIsVllmAvailable:
    """Tests for is_vllm_available function."""

    def test_returns_false_when_imports_fail(self):
        """Should return False when dependencies not installed."""
        with patch.dict("sys.modules", {"vllm": None}):
            # The actual check imports vllm, so we test with import error
            with patch(
                "builtins.__import__",
                side_effect=ImportError("No module named vllm"),
            ):
                # Reset and test - this is tricky due to how imports work
                pass  # The function has try/except so it won't raise


class TestIsLmstudioAvailable:
    """Tests for is_lmstudio_available function (delegates to LMStudioProvider)."""

    def test_returns_true_when_lmstudio_responds(self):
        """Should return True when LMStudioProvider reports available."""
        with patch(
            "local_deep_research.llm.providers.implementations.lmstudio.LMStudioProvider.is_available",
            return_value=True,
        ):
            assert is_lmstudio_available() is True

    def test_returns_false_when_not_running(self):
        """Should return False when LMStudioProvider reports unavailable."""
        with patch(
            "local_deep_research.llm.providers.implementations.lmstudio.LMStudioProvider.is_available",
            return_value=False,
        ):
            assert is_lmstudio_available() is False


class TestIsLlamacppAvailable:
    """Tests for is_llamacpp_available function."""

    def test_returns_true_when_model_path_configured(self):
        """Should return True when model path is configured."""
        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value="/path/to/model.gguf",
        ):
            # Need to also mock the import
            with patch.dict(
                "sys.modules", {"langchain_community.llms": MagicMock()}
            ):
                assert is_llamacpp_available() is True

    def test_returns_false_when_no_model_path(self):
        """Should return False when no model path configured."""
        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value=None,
        ):
            assert is_llamacpp_available() is False


class TestIsGoogleAvailable:
    """Tests for is_google_available function."""

    def test_delegates_to_provider(self):
        """Should delegate to GoogleProvider.is_available."""
        with patch(
            "local_deep_research.config.llm_config.is_google_available"
        ) as mock:
            mock.return_value = True
            # Can't easily test delegation without importing the actual provider
            # Just verify the function exists and can be called
            _result = is_google_available()  # noqa: F841
            # Result depends on actual provider availability


class TestIsOpenrouterAvailable:
    """Tests for is_openrouter_available function."""

    def test_delegates_to_provider(self):
        """Should delegate to OpenRouterProvider.is_available."""
        # Similar to is_google_available
        result = is_openrouter_available()
        assert isinstance(result, bool)


class TestGetAvailableProviders:
    """Tests for get_available_providers function."""

    def test_returns_dict(self):
        """Should return a dictionary."""
        # Clear cache first
        get_available_providers.cache_clear()
        with patch(
            "local_deep_research.config.llm_config.is_ollama_available",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.is_openai_available",
                return_value=False,
            ):
                with patch(
                    "local_deep_research.config.llm_config.is_anthropic_available",
                    return_value=False,
                ):
                    with patch(
                        "local_deep_research.config.llm_config.is_google_available",
                        return_value=False,
                    ):
                        with patch(
                            "local_deep_research.config.llm_config.is_openrouter_available",
                            return_value=False,
                        ):
                            with patch(
                                "local_deep_research.config.llm_config.is_openai_endpoint_available",
                                return_value=False,
                            ):
                                with patch(
                                    "local_deep_research.config.llm_config.is_lmstudio_available",
                                    return_value=False,
                                ):
                                    with patch(
                                        "local_deep_research.config.llm_config.is_llamacpp_available",
                                        return_value=False,
                                    ):
                                        with patch(
                                            "local_deep_research.config.llm_config.is_vllm_available",
                                            return_value=False,
                                        ):
                                            result = get_available_providers()
                                            assert isinstance(result, dict)
                                            # Should have "none" when no providers available
                                            assert "none" in result

    def test_includes_ollama_when_available(self):
        """Should include ollama when available."""
        get_available_providers.cache_clear()
        with patch(
            "local_deep_research.config.llm_config.is_ollama_available",
            return_value=True,
        ):
            with patch(
                "local_deep_research.config.llm_config.is_openai_available",
                return_value=False,
            ):
                with patch(
                    "local_deep_research.config.llm_config.is_anthropic_available",
                    return_value=False,
                ):
                    with patch(
                        "local_deep_research.config.llm_config.is_google_available",
                        return_value=False,
                    ):
                        with patch(
                            "local_deep_research.config.llm_config.is_openrouter_available",
                            return_value=False,
                        ):
                            with patch(
                                "local_deep_research.config.llm_config.is_openai_endpoint_available",
                                return_value=False,
                            ):
                                with patch(
                                    "local_deep_research.config.llm_config.is_lmstudio_available",
                                    return_value=False,
                                ):
                                    with patch(
                                        "local_deep_research.config.llm_config.is_llamacpp_available",
                                        return_value=False,
                                    ):
                                        with patch(
                                            "local_deep_research.config.llm_config.is_vllm_available",
                                            return_value=False,
                                        ):
                                            result = get_available_providers()
                                            assert "ollama" in result


class TestGetSelectedLlmProvider:
    """Tests for get_selected_llm_provider function."""

    def test_returns_provider_from_settings(self):
        """Should return provider from settings."""
        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value="anthropic",
        ):
            result = get_selected_llm_provider()
            assert result == "anthropic"

    def test_returns_lowercase(self):
        """Should return lowercase provider."""
        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value="OPENAI",
        ):
            result = get_selected_llm_provider()
            assert result == "openai"

    def test_defaults_to_ollama(self):
        """Should default to ollama."""
        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value="ollama",
        ) as mock:
            get_selected_llm_provider()
            # Check default is ollama
            mock.assert_called_with(
                "llm.provider", "ollama", settings_snapshot=None
            )


class TestGetFallbackModel:
    """Tests for get_fallback_model function."""

    def test_returns_fake_list_chat_model(self):
        """Should return a FakeListChatModel."""
        from langchain_core.language_models import FakeListChatModel

        result = get_fallback_model()
        assert isinstance(result, FakeListChatModel)

    def test_has_helpful_message(self):
        """Should have a helpful message about no providers."""
        result = get_fallback_model()
        assert len(result.responses) > 0
        assert "No language models" in result.responses[0]


class TestWrapLlmWithoutThinkTags:
    """Tests for wrap_llm_without_think_tags function."""

    def test_returns_wrapper_instance(self):
        """Should return a wrapper instance."""
        mock_llm = MagicMock()
        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value=False,
        ):
            result = wrap_llm_without_think_tags(mock_llm)
            assert hasattr(result, "invoke")
            assert hasattr(result, "base_llm")

    def test_wrapper_invoke_calls_base_llm(self):
        """Should call base LLM on invoke."""
        mock_llm = MagicMock()
        mock_response = MagicMock()
        mock_response.content = "test response"
        mock_llm.invoke.return_value = mock_response

        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value=False,
        ):
            wrapper = wrap_llm_without_think_tags(mock_llm)
            wrapper.invoke("test prompt")
            mock_llm.invoke.assert_called_with("test prompt")

    def test_wrapper_removes_think_tags(self):
        """Should remove think tags from response."""
        mock_llm = MagicMock()
        mock_response = MagicMock()
        mock_response.content = "<think>internal</think>visible"
        mock_llm.invoke.return_value = mock_response

        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.remove_think_tags",
                return_value="visible",
            ) as mock_remove:
                wrapper = wrap_llm_without_think_tags(mock_llm)
                wrapper.invoke("test")
                mock_remove.assert_called_with("<think>internal</think>visible")

    def test_wrapper_delegates_attributes(self):
        """Should delegate attribute access to base LLM."""
        mock_llm = MagicMock()
        mock_llm.model_name = "gpt-4"

        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value=False,
        ):
            wrapper = wrap_llm_without_think_tags(mock_llm)
            assert wrapper.model_name == "gpt-4"

    def test_applies_rate_limiting_when_enabled(self):
        """Should apply rate limiting when enabled in settings."""
        mock_llm = MagicMock()
        mock_wrapped = MagicMock()

        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value=True,
        ):
            # Patch at source location since it's imported inside the function
            with patch(
                "local_deep_research.web_search_engines.rate_limiting.llm.create_rate_limited_llm_wrapper",
                return_value=mock_wrapped,
            ) as mock_create:
                wrap_llm_without_think_tags(mock_llm, provider="openai")
                mock_create.assert_called_with(mock_llm, "openai")


class TestGetLlm:
    """Tests for get_llm function."""

    def test_uses_custom_registered_llm(self):
        """Should use custom LLM when registered."""
        # Import BaseChatModel for proper spec
        from langchain_core.language_models import BaseChatModel

        mock_llm = MagicMock(spec=BaseChatModel)

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=True,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_llm_from_registry",
                return_value=mock_llm,
            ):
                with patch(
                    "local_deep_research.config.llm_config.wrap_llm_without_think_tags",
                    return_value=mock_llm,
                ):
                    with patch(
                        "local_deep_research.config.llm_config.get_setting_from_snapshot",
                        return_value="custom_provider",
                    ):
                        result = get_llm(provider="custom_provider")
                        assert result is mock_llm

    def test_cleans_model_name(self):
        """Should clean model name of quotes and whitespace."""
        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": ' "gpt-4" ',
                    "llm.temperature": 0.7,
                    "llm.provider": "openai",
                }.get(key, default)

                # This will fail trying to create actual LLM, but we're testing name cleaning
                try:
                    get_llm()
                except Exception:
                    pass  # Expected to fail without actual provider

    def test_normalizes_provider_to_lowercase(self):
        """Should normalize provider to lowercase."""
        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "test-model",
                    "llm.temperature": 0.7,
                    "llm.provider": "OPENAI",  # uppercase
                }.get(key, default)

                try:
                    get_llm()
                except Exception:
                    pass  # Expected - testing normalization not actual LLM creation

    def test_invalid_provider_raises_error(self):
        """Should raise ValueError for invalid provider."""
        import pytest

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "test-model",
                    "llm.temperature": 0.7,
                    "llm.provider": "invalid_provider",
                }.get(key, default)

                with pytest.raises(ValueError, match="Invalid provider"):
                    get_llm()

    def test_anthropic_provider_creates_chat_anthropic(self):
        """Should create ChatAnthropic when provider is anthropic."""
        from langchain_anthropic import ChatAnthropic

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "claude-3-sonnet",
                    "llm.temperature": 0.7,
                    "llm.provider": "anthropic",
                    "llm.anthropic.api_key": "sk-ant-test",
                    "llm.local_context_window_size": 4096,
                    "llm.context_window_unrestricted": True,
                    "llm.supports_max_tokens": True,
                    "llm.max_tokens": 4096,
                    "rate_limiting.llm_enabled": False,
                }.get(key, default)

                with patch.object(
                    ChatAnthropic, "__init__", return_value=None
                ) as mock_init:
                    get_llm(provider="anthropic")
                    mock_init.assert_called_once()
                    # Check it was called with model and api_key
                    call_kwargs = mock_init.call_args.kwargs
                    assert call_kwargs["model"] == "claude-3-sonnet"
                    assert call_kwargs["anthropic_api_key"] == "sk-ant-test"

    def test_anthropic_without_api_key_returns_fallback(self):
        """Should return fallback model when anthropic API key not set."""
        from langchain_core.language_models import FakeListChatModel

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "claude-3-sonnet",
                    "llm.temperature": 0.7,
                    "llm.provider": "anthropic",
                    "llm.anthropic.api_key": None,
                    "llm.local_context_window_size": 4096,
                    "llm.context_window_unrestricted": True,
                    "llm.supports_max_tokens": True,
                    "llm.max_tokens": 4096,
                    "rate_limiting.llm_enabled": False,
                }.get(key, default)

                result = get_llm(provider="anthropic")
                # Returns FakeListChatModel directly (not wrapped) when no API key
                assert isinstance(result, FakeListChatModel)

    def test_openai_provider_creates_chat_openai(self):
        """Should create ChatOpenAI when provider is openai."""
        from langchain_openai import ChatOpenAI

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "gpt-4",
                    "llm.temperature": 0.7,
                    "llm.provider": "openai",
                    "llm.openai.api_key": "sk-test",
                    "llm.openai.api_base": None,
                    "llm.openai.organization": None,
                    "llm.streaming": None,
                    "llm.max_retries": None,
                    "llm.request_timeout": None,
                    "llm.local_context_window_size": 4096,
                    "llm.context_window_unrestricted": True,
                    "llm.supports_max_tokens": True,
                    "llm.max_tokens": 4096,
                    "rate_limiting.llm_enabled": False,
                }.get(key, default)

                with patch.object(
                    ChatOpenAI, "__init__", return_value=None
                ) as mock_init:
                    get_llm(provider="openai")
                    mock_init.assert_called_once()
                    call_kwargs = mock_init.call_args.kwargs
                    assert call_kwargs["model"] == "gpt-4"
                    assert call_kwargs["api_key"] == "sk-test"

    def test_openai_without_api_key_returns_fallback(self):
        """Should return fallback model when openai API key not set."""
        from langchain_core.language_models import FakeListChatModel

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "gpt-4",
                    "llm.temperature": 0.7,
                    "llm.provider": "openai",
                    "llm.openai.api_key": None,
                    "llm.local_context_window_size": 4096,
                    "llm.context_window_unrestricted": True,
                    "llm.supports_max_tokens": True,
                    "llm.max_tokens": 4096,
                    "rate_limiting.llm_enabled": False,
                }.get(key, default)

                result = get_llm(provider="openai")
                # Returns FakeListChatModel directly (not wrapped) when no API key
                assert isinstance(result, FakeListChatModel)

    def test_openai_endpoint_creates_chat_openai(self):
        """Should create ChatOpenAI with custom endpoint when provider is openai_endpoint."""
        from langchain_openai import ChatOpenAI

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "custom-model",
                    "llm.temperature": 0.7,
                    "llm.provider": "openai_endpoint",
                    "llm.openai_endpoint.api_key": "custom-key",
                    "llm.openai_endpoint.url": "https://custom.api.com/v1",
                    "llm.local_context_window_size": 4096,
                    "llm.context_window_unrestricted": True,
                    "llm.supports_max_tokens": True,
                    "llm.max_tokens": 4096,
                    "rate_limiting.llm_enabled": False,
                }.get(key, default)

                with patch.object(
                    ChatOpenAI, "__init__", return_value=None
                ) as mock_init:
                    get_llm(provider="openai_endpoint")
                    mock_init.assert_called_once()
                    call_kwargs = mock_init.call_args.kwargs
                    assert call_kwargs["model"] == "custom-model"
                    assert call_kwargs["api_key"] == "custom-key"
                    assert "custom.api.com" in call_kwargs.get(
                        "openai_api_base", ""
                    )

    def test_lmstudio_creates_chat_openai(self):
        """Should create ChatOpenAI with LM Studio URL when provider is lmstudio."""
        from langchain_openai import ChatOpenAI

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "local-model",
                    "llm.temperature": 0.7,
                    "llm.provider": "lmstudio",
                    "llm.lmstudio.url": "http://localhost:1234",
                    "llm.local_context_window_size": 4096,
                    "llm.context_window_unrestricted": False,
                    "llm.context_window_size": 8192,
                    "llm.supports_max_tokens": True,
                    "llm.max_tokens": 4096,
                    "rate_limiting.llm_enabled": False,
                }.get(key, default)

                with patch.object(
                    ChatOpenAI, "__init__", return_value=None
                ) as mock_init:
                    get_llm(provider="lmstudio")
                    mock_init.assert_called_once()
                    call_kwargs = mock_init.call_args.kwargs
                    assert call_kwargs["model"] == "local-model"
                    assert "localhost:1234" in call_kwargs.get("base_url", "")

    def test_custom_factory_function_is_called(self):
        """Should call factory function for custom registered LLM."""
        from langchain_core.language_models import BaseChatModel

        mock_llm = MagicMock(spec=BaseChatModel)
        mock_factory = MagicMock(return_value=mock_llm)

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=True,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_llm_from_registry",
                return_value=mock_factory,
            ):
                with patch(
                    "local_deep_research.config.llm_config.get_setting_from_snapshot"
                ) as mock_get:
                    mock_get.side_effect = lambda key, default=None, **kwargs: {
                        "llm.model": "custom-model",
                        "llm.temperature": 0.5,
                        "llm.provider": "custom_provider",
                        "rate_limiting.llm_enabled": False,
                    }.get(key, default)

                    get_llm(
                        model_name="custom-model",
                        temperature=0.5,
                        provider="custom_provider",
                    )

                    mock_factory.assert_called_once()
                    call_kwargs = mock_factory.call_args.kwargs
                    assert call_kwargs["model_name"] == "custom-model"
                    assert call_kwargs["temperature"] == 0.5

    def test_custom_factory_with_invalid_signature_raises(self):
        """Should raise TypeError when factory has invalid signature."""
        import pytest

        def bad_factory():
            return MagicMock()

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=True,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_llm_from_registry",
                return_value=bad_factory,
            ):
                with patch(
                    "local_deep_research.config.llm_config.get_setting_from_snapshot"
                ) as mock_get:
                    mock_get.side_effect = lambda key, default=None, **kwargs: {
                        "llm.model": "model",
                        "llm.temperature": 0.7,
                        "llm.provider": "bad_factory",
                        "rate_limiting.llm_enabled": False,
                    }.get(key, default)

                    with pytest.raises(TypeError, match="invalid signature"):
                        get_llm(provider="bad_factory")

    def test_custom_factory_returning_non_basechatmodel_raises(self):
        """Should raise ValueError when factory returns non-BaseChatModel."""
        import pytest

        def bad_factory(
            model_name=None, temperature=None, settings_snapshot=None
        ):
            return "not a model"

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=True,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_llm_from_registry",
                return_value=bad_factory,
            ):
                with patch(
                    "local_deep_research.config.llm_config.get_setting_from_snapshot"
                ) as mock_get:
                    mock_get.side_effect = lambda key, default=None, **kwargs: {
                        "llm.model": "model",
                        "llm.temperature": 0.7,
                        "llm.provider": "bad_factory",
                        "rate_limiting.llm_enabled": False,
                    }.get(key, default)

                    with pytest.raises(
                        ValueError, match="must return a BaseChatModel"
                    ):
                        get_llm(provider="bad_factory")

    def test_none_provider_returns_fallback(self):
        """Should return fallback model when provider is 'none'."""
        from langchain_core.language_models import FakeListChatModel

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "test-model",
                    "llm.temperature": 0.7,
                    "llm.provider": "none",
                    "llm.local_context_window_size": 4096,
                    "llm.context_window_unrestricted": True,
                    "llm.supports_max_tokens": True,
                    "llm.max_tokens": 4096,
                    "rate_limiting.llm_enabled": False,
                }.get(key, default)

                result = get_llm(provider="none")
                assert hasattr(result, "base_llm")
                assert isinstance(result.base_llm, FakeListChatModel)

    def test_context_window_size_for_local_providers(self):
        """Should use local context window size for local providers."""
        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                calls = []

                def track_calls(key, default=None, **kwargs):
                    calls.append(key)
                    settings = {
                        "llm.model": "test-model",
                        "llm.temperature": 0.7,
                        "llm.provider": "lmstudio",
                        "llm.lmstudio.url": "http://localhost:1234",
                        "llm.local_context_window_size": 2048,  # Custom local size
                        "llm.context_window_size": 128000,
                        "llm.context_window_unrestricted": False,
                        "llm.supports_max_tokens": True,
                        "llm.max_tokens": 4096,
                        "rate_limiting.llm_enabled": False,
                    }
                    return settings.get(key, default)

                mock_get.side_effect = track_calls

                from langchain_openai import ChatOpenAI

                with patch.object(ChatOpenAI, "__init__", return_value=None):
                    get_llm(provider="lmstudio")
                    # Should have checked local_context_window_size
                    assert "llm.local_context_window_size" in calls

    def test_research_context_is_mutable_parameter(self):
        """Should accept research_context parameter."""

        research_context = {"existing": "value"}

        with patch(
            "local_deep_research.config.llm_config.is_llm_registered",
            return_value=False,
        ):
            with patch(
                "local_deep_research.config.llm_config.get_setting_from_snapshot"
            ) as mock_get:
                mock_get.side_effect = lambda key, default=None, **kwargs: {
                    "llm.model": "test-model",
                    "llm.temperature": 0.7,
                    "llm.provider": "none",  # Falls through to default
                    "llm.local_context_window_size": 4096,
                    "llm.context_window_unrestricted": True,
                    "llm.supports_max_tokens": True,
                    "llm.max_tokens": 4096,
                    "rate_limiting.llm_enabled": False,
                }.get(key, default)

                # Should not fail when research_context is passed
                result = get_llm(
                    provider="none", research_context=research_context
                )
                assert hasattr(result, "base_llm")
                # Original key should still be there
                assert research_context["existing"] == "value"


class TestWrapperStringResponse:
    """Tests for wrapper handling string responses."""

    def test_wrapper_handles_string_response(self):
        """Should handle string response from LLM."""
        mock_llm = MagicMock()
        mock_llm.invoke.return_value = "<think>thought</think>answer"

        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value=False,
        ):
            wrapper = wrap_llm_without_think_tags(mock_llm)
            result = wrapper.invoke("test")
            # Should remove think tags from string
            assert "answer" in result
            assert "<think>" not in result

    def test_wrapper_handles_invoke_exception(self):
        """Should propagate exceptions from LLM invoke."""
        import pytest

        mock_llm = MagicMock()
        mock_llm.invoke.side_effect = RuntimeError("LLM error")

        with patch(
            "local_deep_research.config.llm_config.get_setting_from_snapshot",
            return_value=False,
        ):
            wrapper = wrap_llm_without_think_tags(mock_llm)
            with pytest.raises(RuntimeError, match="LLM error"):
                wrapper.invoke("test")



"""
Deep behavioral tests for state machine patterns.
Tests state transitions, guards, actions,
and workflow state management logic.
"""

from datetime import datetime, timezone


# --- Basic state machine patterns ---


class TestBasicStateMachine:
    """Tests for basic state machine patterns."""

    def _create_state_machine(self, initial_state, transitions):
        """Create a simple state machine."""
        return {
            "current_state": initial_state,
            "transitions": transitions,
            "history": [initial_state],
        }

    def _can_transition(self, machine, event):
        """Check if transition is allowed."""
        current = machine["current_state"]
        transitions = machine["transitions"]
        if current not in transitions:
            return False
        return event in transitions[current]

    def _transition(self, machine, event):
        """Perform state transition."""
        if not self._can_transition(machine, event):
            return False
        current = machine["current_state"]
        new_state = machine["transitions"][current][event]
        machine["current_state"] = new_state
        machine["history"].append(new_state)
        return True

    def test_initial_state(self):
        machine = self._create_state_machine("idle", {})
        assert machine["current_state"] == "idle"

    def test_can_transition(self):
        transitions = {"idle": {"start": "running"}}
        machine = self._create_state_machine("idle", transitions)
        assert self._can_transition(machine, "start") is True

    def test_cannot_transition_invalid_event(self):
        transitions = {"idle": {"start": "running"}}
        machine = self._create_state_machine("idle", transitions)
        assert self._can_transition(machine, "stop") is False

    def test_transition_changes_state(self):
        transitions = {"idle": {"start": "running"}}
        machine = self._create_state_machine("idle", transitions)
        self._transition(machine, "start")
        assert machine["current_state"] == "running"

    def test_history_tracks_states(self):
        transitions = {
            "idle": {"start": "running"},
            "running": {"stop": "stopped"},
        }
        machine = self._create_state_machine("idle", transitions)
        self._transition(machine, "start")
        self._transition(machine, "stop")
        assert machine["history"] == ["idle", "running", "stopped"]


# --- Subscription state patterns ---


class TestSubscriptionStates:
    """Tests for subscription state patterns."""

    def _get_subscription_transitions(self):
        """Get subscription state transitions."""
        return {
            "pending": {"activate": "active", "cancel": "cancelled"},
            "active": {
                "pause": "paused",
                "cancel": "cancelled",
                "expire": "expired",
            },
            "paused": {"resume": "active", "cancel": "cancelled"},
            "expired": {"renew": "active"},
            "cancelled": {},  # Terminal state
        }

    def _is_terminal_state(self, state):
        """Check if state is terminal."""
        transitions = self._get_subscription_transitions()
        return state not in transitions or len(transitions.get(state, {})) == 0

    def _get_available_actions(self, state):
        """Get available actions for state."""
        transitions = self._get_subscription_transitions()
        return list(transitions.get(state, {}).keys())

    def test_pending_can_activate(self):
        actions = self._get_available_actions("pending")
        assert "activate" in actions

    def test_active_can_pause(self):
        actions = self._get_available_actions("active")
        assert "pause" in actions

    def test_cancelled_is_terminal(self):
        assert self._is_terminal_state("cancelled") is True

    def test_active_not_terminal(self):
        assert self._is_terminal_state("active") is False


# --- Card state patterns ---


class TestCardStates:
    """Tests for news card state patterns."""

    def _get_card_transitions(self):
        """Get card state transitions."""
        return {
            "draft": {"publish": "published", "archive": "archived"},
            "published": {"archive": "archived", "feature": "featured"},
            "featured": {"unfeature": "published", "archive": "archived"},
            "archived": {"restore": "published"},
        }

    def _validate_card_state(self, state):
        """Validate card state."""
        valid_states = ["draft", "published", "featured", "archived"]
        return state in valid_states

    def test_draft_can_publish(self):
        transitions = self._get_card_transitions()
        assert "publish" in transitions["draft"]

    def test_published_can_feature(self):
        transitions = self._get_card_transitions()
        assert "feature" in transitions["published"]

    def test_valid_state(self):
        assert self._validate_card_state("published") is True

    def test_invalid_state(self):
        assert self._validate_card_state("unknown") is False


# --- Research state patterns ---


class TestResearchStates:
    """Tests for research task state patterns."""

    def _get_research_transitions(self):
        """Get research state transitions."""
        return {
            "queued": {"start": "running", "cancel": "cancelled"},
            "running": {
                "complete": "completed",
                "fail": "failed",
                "cancel": "cancelled",
            },
            "completed": {"archive": "archived"},
            "failed": {"retry": "queued", "archive": "archived"},
            "cancelled": {"archive": "archived"},
            "archived": {},
        }

    def _can_retry(self, state):
        """Check if research can be retried."""
        return state == "failed"

    def _is_active(self, state):
        """Check if research is active."""
        return state in ["queued", "running"]

    def _is_finished(self, state):
        """Check if research is finished."""
        return state in ["completed", "failed", "cancelled", "archived"]

    def test_failed_can_retry(self):
        assert self._can_retry("failed") is True

    def test_completed_cannot_retry(self):
        assert self._can_retry("completed") is False

    def test_running_is_active(self):
        assert self._is_active("running") is True

    def test_completed_is_finished(self):
        assert self._is_finished("completed") is True

    def test_running_not_finished(self):
        assert self._is_finished("running") is False


# --- Guard condition patterns ---


class TestGuardConditions:
    """Tests for state transition guard conditions."""

    def _check_guard(self, guard_name, context):
        """Check if guard condition passes."""
        guards = {
            "is_authenticated": lambda ctx: ctx.get("user_id") is not None,
            "has_permission": lambda ctx: ctx.get("role")
            in ["admin", "editor"],
            "is_valid": lambda ctx: ctx.get("is_valid", False),
            "not_expired": lambda ctx: ctx.get("expires_at", datetime.max)
            > datetime.now(timezone.utc),
        }
        guard_func = guards.get(guard_name)
        if not guard_func:
            return True
        return guard_func(context)

    def _transition_with_guard(self, machine, event, guard, context):
        """Transition only if guard passes."""
        if not self._check_guard(guard, context):
            return False, "Guard failed"
        # Perform transition logic would go here
        return True, "Success"

    def test_authenticated_guard_passes(self):
        context = {"user_id": "123"}
        assert self._check_guard("is_authenticated", context) is True

    def test_authenticated_guard_fails(self):
        context = {}
        assert self._check_guard("is_authenticated", context) is False

    def test_permission_guard_admin(self):
        context = {"role": "admin"}
        assert self._check_guard("has_permission", context) is True

    def test_permission_guard_user_fails(self):
        context = {"role": "user"}
        assert self._check_guard("has_permission", context) is False


# --- Action patterns ---


class TestStateActions:
    """Tests for state transition actions."""

    def _execute_entry_action(self, state, context):
        """Execute action when entering state."""
        actions = {
            "running": lambda ctx: ctx.update(
                {"started_at": datetime.now(timezone.utc)}
            ),
            "completed": lambda ctx: ctx.update(
                {"completed_at": datetime.now(timezone.utc)}
            ),
            "paused": lambda ctx: ctx.update(
                {"paused_at": datetime.now(timezone.utc)}
            ),
        }
        action = actions.get(state)
        if action:
            action(context)
        return context

    def _execute_exit_action(self, state, context):
        """Execute action when exiting state."""
        actions = {
            "running": lambda ctx: ctx.update(
                {"duration": self._calculate_duration(ctx)}
            ),
        }
        action = actions.get(state)
        if action:
            action(context)
        return context

    def _calculate_duration(self, context):
        """Calculate duration from started_at."""
        started = context.get("started_at")
        if started:
            return (datetime.now(timezone.utc) - started).total_seconds()
        return 0

    def test_entry_action_sets_timestamp(self):
        context = {}
        self._execute_entry_action("running", context)
        assert "started_at" in context

    def test_completed_entry_action(self):
        context = {}
        self._execute_entry_action("completed", context)
        assert "completed_at" in context


# --- Workflow state patterns ---


class TestWorkflowStates:
    """Tests for multi-step workflow states."""

    def _create_workflow(self, steps):
        """Create a workflow with steps."""
        return {
            "steps": steps,
            "current_step": 0,
            "completed_steps": [],
            "status": "pending",
        }

    def _advance_workflow(self, workflow):
        """Advance to next step."""
        if workflow["current_step"] >= len(workflow["steps"]):
            return False
        current = workflow["steps"][workflow["current_step"]]
        workflow["completed_steps"].append(current)
        workflow["current_step"] += 1
        if workflow["current_step"] >= len(workflow["steps"]):
            workflow["status"] = "completed"
        else:
            workflow["status"] = "in_progress"
        return True

    def _get_current_step(self, workflow):
        """Get current step name."""
        if workflow["current_step"] >= len(workflow["steps"]):
            return None
        return workflow["steps"][workflow["current_step"]]

    def _get_progress(self, workflow):
        """Get workflow progress percentage."""
        if not workflow["steps"]:
            return 100
        return (workflow["current_step"] / len(workflow["steps"])) * 100

    def test_workflow_starts_pending(self):
        workflow = self._create_workflow(["step1", "step2"])
        assert workflow["status"] == "pending"

    def test_advance_changes_step(self):
        workflow = self._create_workflow(["step1", "step2"])
        self._advance_workflow(workflow)
        assert workflow["current_step"] == 1

    def test_advance_completes_workflow(self):
        workflow = self._create_workflow(["step1"])
        self._advance_workflow(workflow)
        assert workflow["status"] == "completed"

    def test_progress_calculation(self):
        workflow = self._create_workflow(["a", "b", "c", "d"])
        self._advance_workflow(workflow)
        self._advance_workflow(workflow)
        assert self._get_progress(workflow) == 50


# --- Hierarchical state patterns ---


class TestHierarchicalStates:
    """Tests for hierarchical state machine patterns."""

    def _get_parent_state(self, state, hierarchy):
        """Get parent state from hierarchy."""
        for parent, children in hierarchy.items():
            if state in children:
                return parent
        return None

    def _get_child_states(self, parent, hierarchy):
        """Get child states of a parent."""
        return hierarchy.get(parent, [])

    def _is_in_state(self, current, target, hierarchy):
        """Check if current state is or is within target state."""
        if current == target:
            return True
        parent = self._get_parent_state(current, hierarchy)
        if parent:
            return self._is_in_state(parent, target, hierarchy)
        return False

    def test_get_parent_state(self):
        hierarchy = {"active": ["running", "paused"]}
        parent = self._get_parent_state("running", hierarchy)
        assert parent == "active"

    def test_get_child_states(self):
        hierarchy = {"active": ["running", "paused"]}
        children = self._get_child_states("active", hierarchy)
        assert "running" in children

    def test_is_in_nested_state(self):
        hierarchy = {"active": ["running", "paused"]}
        assert self._is_in_state("running", "active", hierarchy) is True

    def test_not_in_unrelated_state(self):
        hierarchy = {"active": ["running", "paused"]}
        assert self._is_in_state("running", "idle", hierarchy) is False


# --- Parallel state patterns ---


class TestParallelStates:
    """Tests for parallel/orthogonal state patterns."""

    def _create_parallel_machine(self, regions):
        """Create parallel state machine with multiple regions."""
        return {
            "regions": {
                name: {"state": initial} for name, initial in regions.items()
            },
        }

    def _get_region_state(self, machine, region):
        """Get state of a specific region."""
        return machine["regions"].get(region, {}).get("state")

    def _transition_region(self, machine, region, new_state):
        """Transition a specific region."""
        if region in machine["regions"]:
            machine["regions"][region]["state"] = new_state
            return True
        return False

    def _get_combined_state(self, machine):
        """Get combined state from all regions."""
        return {name: r["state"] for name, r in machine["regions"].items()}

    def test_parallel_initial_states(self):
        machine = self._create_parallel_machine(
            {"audio": "muted", "video": "off"}
        )
        assert self._get_region_state(machine, "audio") == "muted"
        assert self._get_region_state(machine, "video") == "off"

    def test_transition_single_region(self):
        machine = self._create_parallel_machine(
            {"audio": "muted", "video": "off"}
        )
        self._transition_region(machine, "audio", "playing")
        assert self._get_region_state(machine, "audio") == "playing"
        assert self._get_region_state(machine, "video") == "off"

    def test_combined_state(self):
        machine = self._create_parallel_machine({"a": "s1", "b": "s2"})
        combined = self._get_combined_state(machine)
        assert combined == {"a": "s1", "b": "s2"}


# --- State persistence patterns ---


class TestStatePersistence:
    """Tests for state persistence patterns."""

    def _serialize_state(self, machine):
        """Serialize state machine to dict."""
        return {
            "current_state": machine["current_state"],
            "history": machine.get("history", []),
            "context": machine.get("context", {}),
        }

    def _deserialize_state(self, data, transitions):
        """Deserialize state machine from dict."""
        return {
            "current_state": data["current_state"],
            "transitions": transitions,
            "history": data.get("history", [data["current_state"]]),
            "context": data.get("context", {}),
        }

    def test_serialize_preserves_state(self):
        machine = {
            "current_state": "running",
            "history": ["idle", "running"],
            "context": {"user": "test"},
        }
        serialized = self._serialize_state(machine)
        assert serialized["current_state"] == "running"
        assert serialized["history"] == ["idle", "running"]

    def test_deserialize_restores_state(self):
        data = {
            "current_state": "paused",
            "history": ["idle", "running", "paused"],
        }
        transitions = {"paused": {"resume": "running"}}
        machine = self._deserialize_state(data, transitions)
        assert machine["current_state"] == "paused"
        assert machine["transitions"] == transitions


# --- State validation patterns ---


class TestStateValidation:
    """Tests for state validation patterns."""

    def _validate_state(self, state, valid_states):
        """Validate state is in valid states."""
        return state in valid_states

    def _validate_transition(self, from_state, to_state, transitions):
        """Validate transition is allowed."""
        if from_state not in transitions:
            return False
        allowed = transitions[from_state].values()
        return to_state in allowed

    def _validate_machine(self, machine):
        """Validate state machine integrity."""
        errors = []
        current = machine.get("current_state")
        transitions = machine.get("transitions", {})
        # Check current state is valid
        all_states = set(transitions.keys())
        for targets in transitions.values():
            all_states.update(targets.values())
        if current not in all_states:
            errors.append(f"Invalid current state: {current}")
        return len(errors) == 0, errors

    def test_valid_state(self):
        assert (
            self._validate_state("active", ["idle", "active", "done"]) is True
        )

    def test_invalid_state(self):
        assert self._validate_state("unknown", ["idle", "active"]) is False

    def test_valid_transition(self):
        transitions = {"idle": {"start": "running"}}
        assert self._validate_transition("idle", "running", transitions) is True

    def test_invalid_transition(self):
        transitions = {"idle": {"start": "running"}}
        assert (
            self._validate_transition("idle", "stopped", transitions) is False
        )


# --- State timeout patterns ---


class TestStateTimeouts:
    """Tests for state timeout patterns."""

    def _check_timeout(self, entered_at, timeout_seconds):
        """Check if state has timed out."""
        if not entered_at:
            return False
        elapsed = (datetime.now(timezone.utc) - entered_at).total_seconds()
        return elapsed > timeout_seconds

    def _get_timeout_for_state(self, state):
        """Get timeout configuration for state."""
        timeouts = {
            "pending": 3600,  # 1 hour
            "processing": 300,  # 5 minutes
            "waiting": 86400,  # 24 hours
        }
        return timeouts.get(state)

    def _should_auto_transition(self, state, entered_at):
        """Check if state should auto-transition due to timeout."""
        timeout = self._get_timeout_for_state(state)
        if not timeout:
            return False
        return self._check_timeout(entered_at, timeout)

    def test_no_timeout_fresh_state(self):
        entered = datetime.now(timezone.utc)
        assert self._check_timeout(entered, 60) is False

    def test_timeout_expired(self):
        from datetime import timedelta

        entered = datetime.now(timezone.utc) - timedelta(seconds=120)
        assert self._check_timeout(entered, 60) is True

    def test_state_has_timeout(self):
        timeout = self._get_timeout_for_state("pending")
        assert timeout == 3600

    def test_state_no_timeout(self):
        timeout = self._get_timeout_for_state("completed")
        assert timeout is None



"""Tests for settings and API key database models."""

from datetime import datetime, timezone

import pytest
from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

from local_deep_research.database.models import (
    APIKey,
    Base,
    Setting,
    SettingType,
    UserSettings,
)


class TestSettingsModels:
    """Test suite for settings-related models."""

    @pytest.fixture
    def engine(self):
        """Create an in-memory SQLite database for testing."""
        engine = create_engine("sqlite:///:memory:")
        Base.metadata.create_all(engine)
        return engine

    @pytest.fixture
    def session(self, engine):
        """Create a database session for testing."""
        Session = sessionmaker(bind=engine)
        session = Session()
        yield session
        session.close()

    def test_setting_creation(self, session):
        """Test creating various types of settings."""
        # App setting
        app_setting = Setting(
            key="app.theme",
            value={"mode": "dark", "accent": "blue"},
            type=SettingType.APP,
            name="Theme",
            category="appearance",
            description="Application theme settings",
            ui_element="select",
            options=["light", "dark", "auto"],
            visible=True,
            editable=True,
        )

        # LLM setting
        llm_setting = Setting(
            key="llm.temperature",
            value=0.7,
            type=SettingType.LLM,
            name="Temperature",
            category="model",
            description="LLM response temperature",
            ui_element="slider",
            min_value=0.0,
            max_value=2.0,
            step=0.1,
            visible=True,
            editable=True,
        )

        # Search setting
        search_setting = Setting(
            key="search.max_results",
            value=20,
            type=SettingType.SEARCH,
            name="Max Results",
            category="search",
            description="Maximum search results per query",
            ui_element="number",
            min_value=1,
            max_value=100,
            visible=True,
            editable=True,
        )

        # Report setting
        report_setting = Setting(
            key="report.format",
            value="markdown",
            type=SettingType.REPORT,
            name="Report Format",
            category="output",
            description="Default report format",
            ui_element="select",
            options=["markdown", "html", "pdf"],
            visible=True,
            editable=True,
        )

        session.add_all(
            [app_setting, llm_setting, search_setting, report_setting]
        )
        session.commit()

        # Verify settings
        all_settings = session.query(Setting).all()
        assert len(all_settings) == 4

        llm = session.query(Setting).filter_by(key="llm.temperature").first()
        assert llm.value == 0.7
        assert llm.type == SettingType.LLM
        assert llm.min_value == 0.0
        assert llm.max_value == 2.0

    def test_unique_key_constraint(self, session):
        """Test that setting keys must be unique."""
        setting1 = Setting(
            key="api.timeout",
            value=30,
            type=SettingType.APP,
            name="API Timeout",
            category="api",
        )

        setting2 = Setting(
            key="api.timeout",  # Duplicate key
            value=60,
            type=SettingType.APP,
            name="API Timeout",
            category="api",
        )

        session.add(setting1)
        session.commit()

        session.add(setting2)
        with pytest.raises(IntegrityError):
            session.commit()

    def test_secret_settings(self, session):
        """Test handling of secret settings (though Setting model doesn't have is_secret)."""
        # Using APIKey for secrets instead
        api_key = APIKey(
            provider="openai",
            key="sk-abc123...",  # Would be encrypted by SQLCipher
            description="OpenAI API key for GPT models",
            is_active=True,
        )

        session.add(api_key)
        session.commit()

        # Verify API key
        saved = session.query(APIKey).filter_by(provider="openai").first()
        assert saved is not None
        assert saved.is_active is True
        assert saved.key == "sk-abc123..."
        # In real usage, the key would be encrypted by SQLCipher

    def test_system_settings(self, session):
        """Test system-wide settings that shouldn't be user-editable."""
        system_setting = Setting(
            key="system.version",
            value="1.0.0",
            type=SettingType.APP,
            name="System Version",
            category="system",
            description="Current system version",
            visible=True,
            editable=False,  # System setting - not editable
        )

        session.add(system_setting)
        session.commit()

        saved = session.query(Setting).filter_by(key="system.version").first()
        assert saved.editable is False
        assert saved.visible is True

    def test_user_settings(self, session):
        """Test UserSettings for user-specific preferences."""
        # User preferences
        preferences = [
            UserSettings(
                key="preferred_model",
                value={"provider": "openai", "model": "gpt-4"},
                category="preferences",
                description="User's preferred LLM model",
            ),
            UserSettings(
                key="search_history_enabled",
                value=True,
                category="privacy",
                description="Whether to save search history",
            ),
            UserSettings(
                key="ui_language",
                value="en",
                category="localization",
                description="User interface language",
            ),
        ]

        session.add_all(preferences)
        session.commit()

        # Query by category
        privacy_settings = (
            session.query(UserSettings).filter_by(category="privacy").all()
        )
        assert len(privacy_settings) == 1
        assert privacy_settings[0].value is True

    def test_api_key_management(self, session):
        """Test APIKey model for secure API key storage."""
        # Add multiple API keys
        keys = [
            APIKey(
                provider="openai",
                key="sk-openai...",
                description="OpenAI GPT models",
                is_active=True,
            ),
            APIKey(
                provider="anthropic",
                key="sk-ant...",
                description="Claude models",
                is_active=True,
            ),
            APIKey(
                provider="google",
                key="AIza...",
                description="Google search API",
                is_active=False,  # Disabled
            ),
        ]

        session.add_all(keys)
        session.commit()

        # Query active keys
        active_keys = session.query(APIKey).filter_by(is_active=True).all()
        assert len(active_keys) == 2

        # Test unique provider constraint
        duplicate = APIKey(
            provider="openai",  # Already exists
            key="sk-different...",
            description="Duplicate provider",
        )

        session.add(duplicate)
        with pytest.raises(IntegrityError):
            session.commit()

    def test_setting_categories(self, session):
        """Test organizing settings by category."""
        categories = ["appearance", "performance", "security", "advanced"]

        for i, cat in enumerate(categories):
            for j in range(3):
                setting = Setting(
                    key=f"{cat}.setting_{j}",
                    value=f"value_{j}",
                    type=SettingType.APP,
                    name=f"{cat.title()} Setting {j}",
                    category=cat,
                    visible=True if i < 3 else False,  # Hide advanced
                )
                session.add(setting)

        session.commit()

        # Query by category
        appearance = (
            session.query(Setting).filter_by(category="appearance").all()
        )
        assert len(appearance) == 3

        # Query visible settings
        visible = session.query(Setting).filter_by(visible=True).count()
        assert visible == 9  # 3 categories * 3 settings each

    def test_setting_value_types(self, session):
        """Test different value types stored in JSON columns."""
        settings = [
            UserSettings(key="string_val", value="text value"),
            UserSettings(key="int_val", value=42),
            UserSettings(key="float_val", value=3.14),
            UserSettings(key="bool_val", value=True),
            UserSettings(key="list_val", value=[1, 2, 3]),
            UserSettings(key="dict_val", value={"a": 1, "b": 2}),
            UserSettings(key="null_val", value=None),
        ]

        session.add_all(settings)
        session.commit()

        # Verify different types
        assert (
            session.query(UserSettings)
            .filter_by(key="string_val")
            .first()
            .value
            == "text value"
        )
        assert (
            session.query(UserSettings).filter_by(key="int_val").first().value
            == 42
        )
        assert (
            session.query(UserSettings).filter_by(key="float_val").first().value
            == 3.14
        )
        assert (
            session.query(UserSettings).filter_by(key="bool_val").first().value
            is True
        )
        assert session.query(UserSettings).filter_by(
            key="list_val"
        ).first().value == [1, 2, 3]
        assert session.query(UserSettings).filter_by(
            key="dict_val"
        ).first().value == {"a": 1, "b": 2}
        assert (
            session.query(UserSettings).filter_by(key="null_val").first().value
            is None
        )

    def test_api_key_rotation(self, session):
        """Test API key rotation and usage tracking."""
        # Create initial key
        api_key = APIKey(
            provider="openai",
            key="sk-old...",
            description="OpenAI API key",
            is_active=True,
            usage_count=100,
        )

        session.add(api_key)
        session.commit()

        # Simulate key usage
        api_key.usage_count += 1
        api_key.last_used = datetime.now(timezone.utc)
        session.commit()

        # Rotate key (deactivate old, add new)
        api_key.is_active = False

        new_key = APIKey(
            provider="openai_new",  # Different provider name to avoid constraint
            key="sk-new...",
            description="OpenAI API key (rotated)",
            is_active=True,
        )

        session.add(new_key)
        session.commit()

        # Verify rotation
        old_key = session.query(APIKey).filter_by(provider="openai").first()
        assert old_key.is_active is False
        assert old_key.usage_count == 101

        active_key = (
            session.query(APIKey)
            .filter_by(is_active=True, provider="openai_new")
            .first()
        )
        assert active_key.key == "sk-new..."

    def test_user_settings_defaults(self, session):
        """Test default values for user settings."""
        # Get a non-existent setting
        setting = (
            session.query(UserSettings)
            .filter_by(key="non_existent_setting")
            .first()
        )

        assert setting is None

        # Create setting with defaults
        default_setting = UserSettings(
            key="new_feature_enabled",
            value=False,  # Default to disabled
            category="features",
            description="New experimental feature",
        )

        session.add(default_setting)
        session.commit()

        saved = (
            session.query(UserSettings)
            .filter_by(key="new_feature_enabled")
            .first()
        )
        assert saved.value is False
        assert saved.created_at is not None
        assert saved.updated_at is not None

    def test_setting_update_tracking(self, session):
        """Test that updated_at is properly tracked."""
        setting = Setting(
            key="test.update",
            value="initial",
            type=SettingType.APP,
            name="Test Update",
            category="test",
        )

        session.add(setting)
        session.commit()

        # Update the setting
        setting.value = "updated"
        session.commit()

        # Note: onupdate might not trigger in SQLite without proper configuration
        # In production with PostgreSQL, this would work automatically



"""
xbench-DeepSearch dataset implementation.

This module provides a class for the xbench-DeepSearch benchmark dataset,
which evaluates deep research and search capabilities.
"""

import base64
from typing import Any, Dict, List, Optional
from loguru import logger
from .base import BenchmarkDataset


class XBenchDeepSearchDataset(BenchmarkDataset):
    """xbench-DeepSearch benchmark dataset for deep research evaluation."""

    @staticmethod
    def xor_decrypt(data: bytes, key: str) -> bytes:
        """XOR decrypt data with a key."""
        key_bytes = key.encode("utf-8")
        key_length = len(key_bytes)
        return bytes(
            [data[i] ^ key_bytes[i % key_length] for i in range(len(data))]
        )

    @classmethod
    def get_dataset_info(cls) -> Dict[str, str]:
        """Get basic information about the dataset."""
        return {
            "id": "xbench_deepsearch",
            "name": "xbench-DeepSearch",
            "description": "Deep research and search capability evaluation (100 questions)",
            "url": "https://huggingface.co/datasets/xbench/DeepSearch",
        }

    @classmethod
    def get_default_dataset_path(cls) -> str:
        """Get the default path for the dataset."""
        return "xbench/DeepSearch"  # Hugging Face dataset identifier

    def load(
        self,
        dataset_path: str = None,
        num_examples: Optional[int] = None,
        seed: int = 42,
    ) -> List[Dict[str, Any]]:
        """Override load to handle HuggingFace datasets directly.

        Args:
            dataset_path: Path to dataset (defaults to HuggingFace)
            num_examples: Optional number of examples to limit
            seed: Random seed for sampling

        Returns:
            List of processed dataset examples
        """
        import random

        # Load the data
        data = self.load_data(dataset_path)

        # Sample if requested
        if num_examples and len(data) > num_examples:
            # Security: seeded random for reproducible benchmark sampling, not security-sensitive
            random.seed(seed)
            data = random.sample(data, num_examples)

        # Process each example
        return [self.process_example(item) for item in data]

    def load_data(
        self,
        dataset_path: str = None,
        num_examples: int = None,
        seed: int = None,
    ) -> List[Dict[str, Any]]:
        """Load the xbench-DeepSearch dataset from Hugging Face.

        Args:
            dataset_path: Path to dataset (defaults to Hugging Face)
            num_examples: Optional number of examples to limit
            seed: Random seed for sampling

        Returns:
            List of questions from xbench-DeepSearch
        """
        try:
            from datasets import load_dataset
        except ImportError:
            logger.exception(
                "datasets library not installed. Run: pip install datasets"
            )
            # Fallback to direct download
            return self._load_from_url(num_examples, seed)

        dataset_path = dataset_path or self.get_default_dataset_path()

        try:
            logger.info(
                f"Loading xbench-DeepSearch dataset from {dataset_path}"
            )

            # Load the dataset from Hugging Face (no authentication needed)
            dataset = load_dataset(dataset_path, split="train")

            # Format for our benchmark system and decrypt
            formatted_questions = []
            for item in dataset:
                # Get the canary key for decryption
                canary = item.get("canary", "")

                # Decrypt prompt and answer if they're encrypted
                prompt = item.get("prompt", "")
                answer = item.get("answer", "")

                try:
                    # Try to decrypt if it looks like base64
                    if prompt and all(
                        c
                        in "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="
                        for c in prompt[:100]
                    ):
                        decrypted_prompt = self.xor_decrypt(
                            base64.b64decode(prompt), canary
                        ).decode("utf-8")
                        prompt = decrypted_prompt

                    if answer and all(
                        c
                        in "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="
                        for c in answer[:100]
                    ):
                        decrypted_answer = self.xor_decrypt(
                            base64.b64decode(answer), canary
                        ).decode("utf-8")
                        answer = decrypted_answer
                except Exception as e:
                    logger.warning(f"Failed to decrypt item: {e}")

                formatted_item = {
                    "id": item.get("id", f"xbench_{len(formatted_questions)}"),
                    "problem": prompt,
                    "answer": answer,
                    "reference_steps": item.get("reference_steps", ""),
                    "canary": canary,
                }
                formatted_questions.append(formatted_item)

            # Apply sampling if num_examples is specified
            if num_examples and num_examples < len(formatted_questions):
                import random

                # Use provided seed or None for random sampling
                # Security: seeded random for reproducible benchmark sampling, not security-sensitive
                if seed is not None:
                    random.seed(seed)
                formatted_questions = random.sample(
                    formatted_questions, num_examples
                )
                logger.info(
                    f"Sampled {num_examples} questions from xbench-DeepSearch (total available: {len(dataset)})"
                )
            else:
                logger.info(
                    f"Loaded {len(formatted_questions)} questions from xbench-DeepSearch"
                )
            return formatted_questions

        except Exception as e:
            logger.warning(f"Failed to load via datasets library: {e}")
            logger.info("Falling back to direct download")
            return self._load_from_url(num_examples, seed)

    def _load_from_url(
        self, num_examples: int = None, seed: int = None
    ) -> List[Dict[str, Any]]:
        """Load dataset directly from URL without datasets library.

        Args:
            num_examples: Optional number of examples to limit
            seed: Random seed for sampling

        Returns:
            List of questions from xbench-DeepSearch
        """
        import pandas as pd

        try:
            # Direct URL to the CSV file on Hugging Face
            url = "https://huggingface.co/datasets/xbench/DeepSearch/resolve/main/data/train-00000-of-00001.parquet"

            logger.info(f"Downloading xbench-DeepSearch from {url}")
            df = pd.read_parquet(url)

            # Convert to list of dicts and decrypt
            questions = []
            for _, row in df.iterrows():
                # Get the canary key for decryption
                canary = row.get("canary", "")

                # Decrypt prompt and answer
                prompt = row.get("prompt", "")
                answer = row.get("answer", "")

                try:
                    if prompt and all(
                        c
                        in "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="
                        for c in prompt[:100]
                    ):
                        decrypted_prompt = self.xor_decrypt(
                            base64.b64decode(prompt), canary
                        ).decode("utf-8")
                        prompt = decrypted_prompt

                    if answer and all(
                        c
                        in "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="
                        for c in answer[:100]
                    ):
                        decrypted_answer = self.xor_decrypt(
                            base64.b64decode(answer), canary
                        ).decode("utf-8")
                        answer = decrypted_answer
                except Exception as e:
                    logger.warning(f"Failed to decrypt item: {e}")

                questions.append(
                    {
                        "id": row.get("id", f"xbench_{len(questions)}"),
                        "problem": prompt,
                        "answer": answer,
                        "reference_steps": row.get("reference_steps", ""),
                        "canary": canary,
                    }
                )

            # Apply sampling if num_examples is specified
            if num_examples and num_examples < len(questions):
                import random

                # Use provided seed or None for random sampling
                # Security: seeded random for reproducible benchmark sampling, not security-sensitive
                if seed is not None:
                    random.seed(seed)
                questions = random.sample(questions, num_examples)
                logger.info(
                    f"Sampled {num_examples} questions via direct download (total available: {len(questions)})"
                )
            else:
                logger.info(
                    f"Loaded {len(questions)} questions via direct download"
                )
            return questions

        except Exception:
            logger.exception("Failed to load dataset")
            return []

    def process_example(self, example: Dict[str, Any]) -> Dict[str, Any]:
        """Process a single example from the dataset.

        xbench-DeepSearch questions are designed for deep research evaluation.
        """
        processed = dict(example)

        # Add evaluation metadata
        processed["requires_deep_search"] = True
        processed["expected_iterations"] = (
            4  # Deep search questions need multiple iterations
        )

        # Evaluation criteria for research questions
        processed["evaluation_criteria"] = {
            "accuracy": 0.4,
            "completeness": 0.3,
            "reasoning": 0.2,
            "sources": 0.1,  # Credit for citing sources
        }

        return processed



"""Tests for pagination bounds enforcement (PR #1956).

Verifies:
- History routes clamp limit/offset/page/per_page to safe ranges
- Context overflow time series capped at 1000 for short periods
"""

import inspect


class TestHistoryRoutesPagination:
    """Verify get_history enforces limit/offset bounds."""

    def _get_source(self):
        from local_deep_research.web.routes.history_routes import get_history

        return inspect.getsource(get_history)

    def test_limit_clamped_to_max_500(self):
        """Source should clamp limit to max 500."""
        source = self._get_source()
        assert "min(limit, 500)" in source or "min(limit,500)" in source

    def test_limit_clamped_to_min_1(self):
        """Source should clamp limit to min 1."""
        source = self._get_source()
        assert "max(1," in source

    def test_offset_clamped_to_min_0(self):
        """Source should clamp offset to min 0."""
        source = self._get_source()
        assert "max(0, offset)" in source or "max(0,offset)" in source


class TestResearchRoutesPagination:
    """Verify api_get_history enforces page/per_page bounds."""

    def test_source_clamps_per_page_to_500(self):
        """per_page should be clamped to max 500."""
        from local_deep_research.web.routes.research_routes_orm import (
            api_get_history,
        )

        source = inspect.getsource(api_get_history)
        assert "min(per_page, 500)" in source or "min(per_page,500)" in source

    def test_source_clamps_page_min_1(self):
        """page should be clamped to min 1."""
        from local_deep_research.web.routes.research_routes_orm import (
            api_get_history,
        )

        source = inspect.getsource(api_get_history)
        assert "max(1," in source

    def test_source_clamps_per_page_min_1(self):
        """per_page should be clamped to min 1."""
        from local_deep_research.web.routes.research_routes_orm import (
            api_get_history,
        )

        source = inspect.getsource(api_get_history)
        assert "max(1, min(per_page" in source


class TestContextOverflowCap:
    """Verify time series query has a limit cap."""

    def test_short_period_capped_at_1000(self):
        """Short period time series queries should be capped at 1000."""
        from local_deep_research.web.routes.context_overflow_api import (
            get_context_overflow_metrics,
        )

        source = inspect.getsource(get_context_overflow_metrics)
        assert "limit(1000)" in source



"""Tests for search cache persistence and caching edge cases.

These tests exercise the SearchCache's persistence across instances,
the get_or_fetch lifecycle with falsy results, and JSON serialization
roundtrips through SQLite.
"""

import pytest

from local_deep_research.utilities.search_cache import SearchCache


@pytest.fixture
def cache_dir(tmp_path):
    """Provide a temporary cache directory."""
    return str(tmp_path / "test_cache")


@pytest.fixture
def cache(cache_dir):
    """Provide a SearchCache instance with a temporary directory."""
    c = SearchCache(cache_dir=cache_dir)
    yield c
    c.dispose()


class TestPersistenceAcrossInstances:
    """Tests for data surviving across cache instance lifecycles."""

    def test_persistence_across_cache_instances(self, cache_dir):
        """Data put with instance 1 survives after dispose and re-creation.
        This is the fundamental purpose of the persistent cache."""
        cache1 = SearchCache(cache_dir=cache_dir)
        results = [{"title": "Persistent Result", "url": "https://example.com"}]
        cache1.put("persistent query", results, "test_engine")
        cache1.dispose()

        # Create a new instance with the same directory
        cache2 = SearchCache(cache_dir=cache_dir)
        retrieved = cache2.get("persistent query", "test_engine")
        cache2.dispose()

        assert retrieved is not None
        assert len(retrieved) == 1
        assert retrieved[0]["title"] == "Persistent Result"


class TestGetOrFetchWithFalsyResults:
    """Tests for get_or_fetch behavior with empty/None results."""

    def test_empty_list_result_not_cached_via_get_or_fetch(self, cache):
        """fetch_func returns [] (falsy) — it's returned to caller but NOT
        cached, causing re-fetch on next call. This documents the behavior
        where put() rejects empty results."""
        call_count = 0

        def fetch_empty():
            nonlocal call_count
            call_count += 1
            return []

        # First call: fetches and returns []
        result1 = cache.get_or_fetch("empty query", fetch_empty, "test_engine")
        assert result1 == []
        assert call_count == 1

        # Clear memory cache to force DB lookup
        cache._memory_cache.clear()

        # Second call: fetches AGAIN because [] wasn't cached
        result2 = cache.get_or_fetch("empty query", fetch_empty, "test_engine")
        assert result2 == []
        assert (
            call_count == 2
        )  # Called again because empty results aren't cached

    def test_fetch_returning_none_path(self, cache):
        """fetch_func returning None (not exception) goes through the full
        stampede protection path and returns None."""

        def fetch_none():
            return None

        result = cache.get_or_fetch("none query", fetch_none, "test_engine")
        assert result is None

    def test_fetch_exception_returns_none(self, cache):
        """fetch_func raising an exception returns None (not the exception)."""

        def fetch_error():
            raise RuntimeError("network error")

        result = cache.get_or_fetch("error query", fetch_error, "test_engine")
        assert result is None


class TestMemoryCacheMasksDbAccess:
    """Tests documenting memory cache interaction with DB stats."""

    def test_memory_cache_masks_db_access_count(self, cache):
        """Memory cache satisfies reads, so DB access_count never increments
        beyond the initial put. This documents a stats limitation."""
        results = [{"title": "Test", "url": "https://example.com"}]
        cache.put("stats query", results, "test_engine")

        # Multiple reads from memory cache
        for _ in range(5):
            cache.get("stats query", "test_engine")

        # Check DB access_count - should be 1 (from initial put)
        # because all reads were served from memory cache
        with cache.Session() as session:
            from local_deep_research.database.models import (
                SearchCache as SearchCacheModel,
            )

            query_hash = cache._get_query_hash("stats query", "test_engine")
            entry = (
                session.query(SearchCacheModel)
                .filter_by(query_hash=query_hash)
                .first()
            )
            assert entry is not None
            # access_count stays at 1 because memory cache serves all reads
            assert entry.access_count == 1


class TestJsonSerializationRoundtrip:
    """Tests for complex data surviving SQLite JSON serialization."""

    def test_complex_nested_results_json_roundtrip(self, cache_dir):
        """Deeply nested structures with None, unicode, and special chars
        survive the SQLite JSON serialization roundtrip."""
        complex_results = [
            {
                "title": "Unicode Test: \u00e4\u00f6\u00fc \u00df \u2603 \u2764",
                "url": "https://example.com/path?q=test&lang=en",
                "metadata": {
                    "nested": {
                        "deep": {
                            "value": 42,
                            "list": [1, None, "three", True, False],
                        }
                    },
                    "empty_dict": {},
                    "empty_list": [],
                    "null_value": None,
                },
                "snippet": 'Text with "quotes" and\nnewlines\tand\ttabs',
            }
        ]

        cache = SearchCache(cache_dir=cache_dir)
        cache.put("complex query", complex_results, "test_engine")
        cache.dispose()

        # Retrieve from a fresh instance (forces DB read)
        cache2 = SearchCache(cache_dir=cache_dir)
        retrieved = cache2.get("complex query", "test_engine")
        cache2.dispose()

        assert retrieved is not None
        assert len(retrieved) == 1
        result = retrieved[0]
        assert "\u00e4\u00f6\u00fc" in result["title"]
        assert result["metadata"]["nested"]["deep"]["value"] == 42
        assert result["metadata"]["nested"]["deep"]["list"] == [
            1,
            None,
            "three",
            True,
            False,
        ]
        assert result["metadata"]["null_value"] is None
        assert "\n" in result["snippet"]



"""
Tests for the LLM provider auto-discovery system.

Tests cover:
- ProviderInfo class
- ProviderDiscovery singleton
- Provider discovery and registration
- Getting provider options for UI
"""

from unittest.mock import Mock


class TestProviderInfo:
    """Tests for ProviderInfo class."""

    def test_init_extracts_provider_key(self):
        """ProviderInfo extracts provider key from class."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderInfo,
        )

        mock_class = Mock()
        mock_class.__name__ = "TestProvider"
        mock_class.provider_name = "Test"
        mock_class.provider_key = "TEST"
        mock_class.company_name = "Test Company"
        mock_class.is_cloud = True
        mock_class.requires_auth_for_models = Mock(return_value=True)

        info = ProviderInfo(mock_class)

        assert info.provider_key == "TEST"
        assert info.provider_name == "Test"

    def test_init_generates_provider_key_from_class_name(self):
        """ProviderInfo generates key from class name if not specified."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderInfo,
        )

        # Create a mock class without provider_key attribute
        mock_class = Mock()
        mock_class.__name__ = "CustomProvider"
        mock_class.provider_name = "Custom"
        # Use spec to control which attributes exist
        del (
            mock_class.provider_key
        )  # Remove provider_key so it falls back to class name
        mock_class.company_name = "Custom Company"
        mock_class.is_cloud = True
        mock_class.requires_auth_for_models = Mock(return_value=True)

        info = ProviderInfo(mock_class)

        assert info.provider_name == "Custom"

    def test_to_dict_returns_expected_keys(self):
        """to_dict returns expected dictionary keys."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderInfo,
        )

        mock_class = Mock()
        mock_class.__name__ = "TestProvider"
        mock_class.provider_name = "Test Provider"
        mock_class.provider_key = "TEST"
        mock_class.company_name = "Test Company"
        mock_class.is_cloud = True
        mock_class.requires_auth_for_models = Mock(return_value=True)

        info = ProviderInfo(mock_class)
        result = info.to_dict()

        assert "value" in result
        assert "label" in result
        assert "is_cloud" in result

    def test_generate_display_name_local_provider(self):
        """Display name shows local indicator for non-cloud providers."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderInfo,
        )

        mock_class = Mock()
        mock_class.__name__ = "LocalProvider"
        mock_class.provider_name = "Local LLM"
        mock_class.provider_key = "LOCAL"
        mock_class.company_name = "Local Company"
        mock_class.is_cloud = False
        mock_class.requires_auth_for_models = Mock(return_value=False)

        info = ProviderInfo(mock_class)

        # Display name should indicate local
        assert "Local" in info.display_name


class TestProviderDiscovery:
    """Tests for ProviderDiscovery class."""

    def test_singleton_pattern(self):
        """ProviderDiscovery follows singleton pattern."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
        )

        instance1 = ProviderDiscovery()
        instance2 = ProviderDiscovery()

        assert instance1 is instance2

    def test_discover_providers_returns_dict(self):
        """discover_providers returns dictionary."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
        )

        discovery = ProviderDiscovery()
        result = discovery.discover_providers()

        assert isinstance(result, dict)

    def test_get_provider_options_returns_list(self):
        """get_provider_options returns list."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
        )

        discovery = ProviderDiscovery()
        result = discovery.get_provider_options()

        assert isinstance(result, list)

    def test_get_provider_options_sorted_by_label(self):
        """get_provider_options returns sorted list."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
        )

        discovery = ProviderDiscovery()
        options = discovery.get_provider_options()

        if len(options) > 1:
            labels = [opt["label"] for opt in options]
            assert labels == sorted(labels)

    def test_get_provider_info_returns_provider(self):
        """get_provider_info returns ProviderInfo for known provider."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
        )

        discovery = ProviderDiscovery()
        discovery.discover_providers()

        # Should have discovered some providers
        if discovery._providers:
            # Get first provider key
            first_key = list(discovery._providers.keys())[0]
            info = discovery.get_provider_info(first_key)

            assert info is not None

    def test_get_provider_info_returns_none_for_unknown(self):
        """get_provider_info returns None for unknown provider."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
        )

        discovery = ProviderDiscovery()
        result = discovery.get_provider_info("NONEXISTENT_PROVIDER_XYZ")

        assert result is None

    def test_get_provider_class_returns_class(self):
        """get_provider_class returns provider class."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
        )

        discovery = ProviderDiscovery()
        discovery.discover_providers()

        # Should have discovered some providers
        if discovery._providers:
            first_key = list(discovery._providers.keys())[0]
            cls = discovery.get_provider_class(first_key)

            assert cls is not None

    def test_get_provider_class_returns_none_for_unknown(self):
        """get_provider_class returns None for unknown provider."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
        )

        discovery = ProviderDiscovery()
        result = discovery.get_provider_class("NONEXISTENT_PROVIDER_XYZ")

        assert result is None


class TestModuleFunctions:
    """Tests for module-level functions."""

    def test_discover_providers_function(self):
        """discover_providers function works."""
        from local_deep_research.llm.providers.auto_discovery import (
            discover_providers,
        )

        result = discover_providers()

        assert isinstance(result, dict)

    def test_get_discovered_provider_options_function(self):
        """get_discovered_provider_options function works."""
        from local_deep_research.llm.providers.auto_discovery import (
            get_discovered_provider_options,
        )

        result = get_discovered_provider_options()

        assert isinstance(result, list)

    def test_get_provider_class_function(self):
        """get_provider_class function works."""
        from local_deep_research.llm.providers.auto_discovery import (
            get_provider_class,
        )

        # Test with unknown provider
        result = get_provider_class("UNKNOWN_XYZ")

        assert result is None


class TestGlobalInstance:
    """Tests for global provider_discovery instance."""

    def test_global_instance_exists(self):
        """Global provider_discovery instance exists."""
        from local_deep_research.llm.providers.auto_discovery import (
            provider_discovery,
        )

        assert provider_discovery is not None

    def test_global_instance_is_provider_discovery(self):
        """Global instance is ProviderDiscovery type."""
        from local_deep_research.llm.providers.auto_discovery import (
            ProviderDiscovery,
            provider_discovery,
        )

        assert isinstance(provider_discovery, ProviderDiscovery)
